import errno
import json
import os
from argparse import Namespace
from configparser import ConfigParser, NoSectionError
from typing import List, Set, Any, Callable, Optional

from wordfence.logging import log
from .config_items import Context, ConfigItemDefinition, \
    CanonicalValueExtractorInterface, not_set_token, \
    ReferenceToken, merge_config_maps
from .defaults import INI_DEFAULT_PATH
from .base_config_definitions import config_map as base_config_map
from ..subcommands import SubcommandDefinition


valid_contexts: Set[Context] = {Context.ALL, Context.CONFIG}


GLOBAL_INI_PATH = b'/etc/wordfence/wordfence-cli.ini'
DEFAULT_SECTION_NAME = 'DEFAULT'


class IniCanonicalValueExtractor(CanonicalValueExtractorInterface):

    def __init__(self, *config_section_names):
        self.config_section_names = config_section_names

    def is_valid_source(self, source: Any) -> bool:
        return isinstance(source, ConfigParser)

    def _get_value_from_section(
                self,
                definition: ConfigItemDefinition,
                source: ConfigParser,
                section: str
            ) -> Any:
        if definition.has_separator():
            # always return separated values as a string
            return source.get(
                    section,
                    definition.property_name,
                    fallback=not_set_token
                )
        elif definition.get_value_type() == bool:
            return source.getboolean(
                    section,
                    definition.property_name,
                    fallback=not_set_token
                )
        elif definition.get_value_type() == int:
            return source.getint(
                    section,
                    definition.property_name,
                    fallback=not_set_token
                )
        elif definition.accepts_paths():
            path = source.get(
                    section,
                    definition.property_name,
                    fallback=not_set_token
                )
            if path != not_set_token:
                path = os.fsencode(path)
            return path
        elif isinstance(definition.get_value_type(), Callable):
            value = source.get(
                    section,
                    definition.property_name,
                    fallback=not_set_token
                )
            # convert using the type method
            value = value if isinstance(value, ReferenceToken) else (
                (definition.get_value_type())(value))
            return value
        elif definition.get_value_type() != str:
            raise ValueError(
                    "Only string, bool, int, and callable types are currently "
                    "known to the INI parser"
                )
        else:
            return source.get(
                    section,
                    definition.property_name,
                    fallback=not_set_token
                )

    def get_canonical_value(self, definition: ConfigItemDefinition,
                            source: ConfigParser) -> Any:
        self.assert_is_valid_source(source)

        for section in self.config_section_names:
            value = self._get_value_from_section(
                    definition,
                    source,
                    section
                )
            if value is not None and value is not not_set_token:
                break

        if isinstance(value, str) and definition.has_separator():
            value = value.split(definition.meta.separator)
            if definition.get_value_type() == int:
                value = [int(string_int) for string_int in value]
            elif definition.get_value_type() != str:
                raise ValueError(
                    "INI files currently support lists of strings and ints, no"
                    " other types")

        return value

    def get_context(self) -> Context:
        return Context.CONFIG


def get_ini_value_extractor(
            subcommand_definition: SubcommandDefinition
        ) -> IniCanonicalValueExtractor:
    return IniCanonicalValueExtractor(
            subcommand_definition.config_section,
            DEFAULT_SECTION_NAME
        )


def get_default_ini_value_extractor() -> IniCanonicalValueExtractor:
    return IniCanonicalValueExtractor(DEFAULT_SECTION_NAME)


def get_ini_path(cli_values: Namespace) -> str:
    if 'configuration' not in cli_values or not isinstance(
            cli_values.configuration, bytes):
        path = INI_DEFAULT_PATH
    else:
        path = cli_values.configuration
    return os.path.expanduser(path)


def load_ini(
            cli_values,
            subcommand_definition: Optional[SubcommandDefinition]
        ) -> (ConfigParser, Optional[str]):
    config = ConfigParser()
    try:
        with open(GLOBAL_INI_PATH, 'r') as file:
            config.read_file(file)
    except FileNotFoundError:
        pass  # Ignore nonexistant global config files
    except OSError:
        log.warning(f'Failed to read global config file at {GLOBAL_INI_PATH}')
    ini_path = get_ini_path(cli_values)
    try:
        with open(ini_path) as file:
            config.read_file(file)
    except OSError as e:
        if e.errno == errno.EACCES:
            raise PermissionError(
                f"The current user cannot read the config file: "
                f"{json.dumps(get_ini_path(cli_values))}") from e
        elif e.errno != errno.ENOENT:
            raise
        # config file does not exist: proceed with default values + CLI values
        return (config, None)
    section_map = {
            DEFAULT_SECTION_NAME: base_config_map,
        }
    if subcommand_definition is not None:
        section_map[subcommand_definition.config_section] = \
                merge_config_maps(
                        base_config_map,
                        subcommand_definition.get_config_map()
                    )
    all_section_names: List[str] = config.sections()
    invalid_settings: bool = False
    for section_name in all_section_names:
        if section_name not in section_map:
            config.remove_section(section_name)
    # remove values that are in the incorrect context or are entirely unknown
    for section, definitions in section_map.items():
        try:
            items = config.items(section)
        except NoSectionError:
            items = {}
        for property_name, _value in items:
            # arguments are stored in the lookup by name (kebab-case), but
            # written out in snake_case in the INI
            key = property_name.replace('_', '-')
            # detect unknown definitions and definitions written in kebab-case
            # instead of snake_case
            if key not in definitions or (
                    key == property_name and '-' in property_name):
                log.warning(
                        "Ignoring unknown config setting "
                        f"{json.dumps(property_name)}"
                    )
                config.remove_option(section, property_name)
                invalid_settings = True
            if key in definitions:
                valid_ini_value = definitions[key].context in valid_contexts
                if not valid_ini_value:
                    log.warning(
                        f"Ignoring setting that is not valid in the config "
                        f"file context: {json.dumps(definitions[key].name)}.")
                    invalid_settings = True
                    config.remove_option(section, property_name)
                    continue
    if invalid_settings:
        log.warning(
            "*** Invalid settings not known to wordfence-cli or that are not "
            "intended for use in INI config files were discarded. ***")

    return (config, ini_path)



from html import escape
from typing import Dict, Optional, List, Union, Any


class HtmlContent:

    def to_html(self) -> str:
        raise NotImplementedError(
                'HTML content must be able to be converted to a string'
            )

    def __str__(self) -> str:
        return self.to_html()


class RawHtml(HtmlContent):

    def __init__(self, html: str):
        self.html = html

    def to_html(self) -> str:
        return self.html


def to_html(
            content: Optional[List[Union[HtmlContent, str]]] = None
        ) -> str:
    string = ''
    for item in content:
        if isinstance(item, HtmlContent):
            string += item.to_html()
        else:
            string += escape(item)
    return string


class Container(HtmlContent):

    def __init__(
                self,
                content: Optional[List[Union[HtmlContent, str]]] = None
            ):
        self.content = content if content is not None else []

    def append(self, content: Any):
        if not (isinstance(content, HtmlContent) or isinstance(content, str)):
            content = str(content)
        self.content.append(content)
        return self

    def to_html(self) -> str:
        return to_html(self.content)


class Tag(Container):

    def __init__(
                self,
                name: str,
                attributes: Optional[Dict[str, Optional[str]]] = None,
                content: Optional[List[Union[HtmlContent, str]]] = None
            ):
        self.name = name
        self.attributes = attributes if attributes is not None else {}
        super().__init__(content)

    def set_attribute(self, name: str, value: Optional[str] = None):
        self.attributes[name] = value
        return self

    def _format_attributes(self) -> str:
        attribute_string = ''
        for name, value in self.attributes.items():
            name = escape(name)
            value = escape(value)
            attribute_string += f' {name}="{value}"'
        return attribute_string

    def to_html(self) -> str:
        attribute_string = self._format_attributes()
        string = f'<{self.name}{attribute_string}>'
        if len(self.content) > 0:
            string += super().to_html()
            string += f'</{self.name}>'
        return string


class Document(HtmlContent):

    def __init__(self):
        self.head = Tag('head')
        self.body = Tag('body')

    def to_html(self) -> str:
        html = Tag('html')
        html.append(self.head)
        html.append(self.body)
        return html.to_html()


class Style:

    def __init__(
                self,
                selector: str,
                properties: Optional[Dict[str, str]] = None
            ):
        self.selector = selector
        self.properties = properties if properties is not None else {}

    def set(self, property: str, value):
        self.properties[property] = str(value)
        return self

    def __str__(self) -> str:
        css = self.selector + ' {\n'
        for name, value in self.properties.items():
            css += f'\t{name}: {value};\n'
        css += '}'
        return css


class Stylesheet(HtmlContent):

    def __init__(
                self,
                styles: Optional[List[Style]] = None
            ):
        self.styles = styles if styles is not None else []

    def add(self, *styles: Style):
        self.styles.extend(styles)
        return self

    def to_html(self) -> str:
        tag = Tag('style')
        for style in self.styles:
            tag.append(str(style))
        return tag.to_html()



import unittest

from .versioning import compare_php_versions


class TestPhpVersions(unittest.TestCase):

    def _expect_comparison(
                self,
                a: str,
                b: str,
                expected: int
            ) -> None:
        self.assertEqual(
                compare_php_versions(a, b),
                expected
            )

    def test_numeric(self):
        self._expect_comparison('1.0.0', '1.0.0', 0)
        self._expect_comparison('2.0.0', '3.0.0', -1)
        self._expect_comparison('5.0.0', '4.0.0', 1)

    def test_short(self):
        self._expect_comparison('1.0', '1.0.0', 0)
        self._expect_comparison('2.0.0', '2', 0)
        self._expect_comparison('1', '2', -1)

    def test_dev_versions(self):
        self._expect_comparison('1.0.0-rc1', '1.0.0', -1)
        self._expect_comparison('2.0.dev', '2.0', -1)
        self._expect_comparison('1.0.0-test', '1.0.0-dev', -1)
        self._expect_comparison('1.0.0-alpha', '1.0.0-dev', 1)
        self._expect_comparison('3.45.beta', '3.45.alpha', 1)
        self._expect_comparison('5.0.0.beta', '5.0.0.rc', -1)
        self._expect_comparison('9.0.0', '9.0.0rc1', 1)
        self._expect_comparison('10.0.0', '10.0.0pl', -1)
        self._expect_comparison('1.0-dev', '1.0.0-dev', -1)
        self._expect_comparison('1.0.0-rc', '1.0.0-RC', 0)
        self._expect_comparison('1.0.0-alpha', '1.0.0-a', 0)
        self._expect_comparison('1.0.0-beta', '1.0.0b', 0)
        self._expect_comparison('1.0.0-pl', '1.0.0-p', 0)



from typing import Optional


class ApiException(Exception):

    def __init__(
                self,
                internal_message: str,
                public_message: Optional[str] = None
            ):
        if public_message is not None:
            message = f'{internal_message}: {public_message}'
        else:
            message = internal_message
        super().__init__(message)
        self.public_message = public_message



from typing import Union, Generator, List
from wordfence.intel.database_rules import DatabaseRuleSet, DatabaseRule
from wordfence.wordpress.database import WordpressDatabase, \
    WordpressDatabaseConnection
from wordfence.logging import log, VERBOSE
from wordfence.util.timing import Timer


class DatabaseScanResult:

    def __init__(
                self,
                rule: DatabaseRule,
                table: str,
                row: dict
            ):
        self.rule = rule
        self.table = table
        self.row = row


class DatabaseScanner:

    def __init__(
                self,
                rule_set: DatabaseRuleSet
            ):
        self.rule_set = rule_set
        self.scan_count = 0
        self.timer = Timer(start=False)

    def _get_valid_columns(
                self,
                connection: WordpressDatabaseConnection,
                prefixed_table: str
            ) -> List:
        columns = connection.get_column_types(prefixed_table)
        try:
            del columns['rule_id']
        except KeyError:
            pass  # If the column doesn't exist, that's fine
        return list(columns.keys())

    def _scan_table(
                self,
                connection: WordpressDatabaseConnection,
                table: str
            ) -> Generator[DatabaseScanResult, None, None]:
        prefixed_table = connection.prefix_table(table)
        conditions = []
        rule_selects = []
        for rule in self.rule_set.get_rules(table):
            conditions.append(f'({rule.condition})')
            rule_selects.append(
                    f'WHEN {rule.condition} THEN {rule.identifier}'
                )
        rule_case = 'CASE\n' + '\n'.join(rule_selects) + '\nEND'
        selected_columns = self._get_valid_columns(connection, prefixed_table)
        selected_columns.append(f'{rule_case} as rule_id')
        selected_columns = ', '.join(selected_columns)
        query = (
                f'SELECT {selected_columns} FROM '
                f'{prefixed_table} WHERE '
                + ' OR '.join(conditions)
            )
        # Using a dict as the query parameters avoids %s from being
        # interpreted as a placeholder (there is apparently no way
        # to escape "%s" ("%%s" doesn't work)
        for result in connection.query_literal(query):
            rule = self.rule_set.get_rule(result['rule_id'])
            del result['rule_id']
            yield DatabaseScanResult(
                    rule=rule,
                    table=prefixed_table,
                    row=result
                )

    def _scan_connection(
                self,
                connection: WordpressDatabaseConnection
            ) -> Generator[DatabaseScanResult, None, None]:
        self.timer.resume()
        log.log(
                VERBOSE,
                f'Scanning database: {connection.database.debug_string}...'
            )
        for table in self.rule_set.get_targeted_tables():
            yield from self._scan_table(connection, table)
        log.log(
                VERBOSE,
                f'Scan completed for: {connection.database.debug_string}'
            )
        self.timer.stop()

    def scan(
                self,
                database: Union[WordpressDatabase, WordpressDatabaseConnection]
            ) -> Generator[DatabaseScanResult, None, None]:
        self.scan_count += 1
        if isinstance(database, WordpressDatabaseConnection):
            yield from self._scan_connection(database)
        else:
            log.log(
                    VERBOSE,
                    f'Connecting to database: {database.debug_string}...'
                )
            with database.connect() as connection:
                log.log(
                        VERBOSE,
                        'Successfully connected to database: '
                        f'{database.debug_string}'
                    )
                yield from self._scan_connection(connection)

    def get_elapsed_time(self) -> int:
        return self.timer.get_elapsed()



from ctypes import cdll, CDLL
from ctypes.util import find_library
from importlib import import_module


class LibraryNotAvailableException(Exception):
    pass


def load_library(name: str) -> CDLL:
    pathname = find_library(name)
    if pathname is None:
        raise LibraryNotAvailableException()
    try:
        library = cdll.LoadLibrary(pathname)
        return library
    except OSError:
        raise LibraryNotAvailableException


class OptionalUtility:

    def __init__(self, name):
        try:
            self.module = import_module('.' + name, 'wordfence.util')
        except LibraryNotAvailableException:
            self.module = None

    def is_available(self) -> bool:
        return self.module is not None



import re

from collections import deque
from enum import Enum, auto
from typing import Generator, BinaryIO, Optional, Union, Set

from wordfence.util.encoding import bytes_to_str


class LexingException(Exception):
    pass


class MatchType(Enum):
    NONE = auto(),
    MATCH = auto(),
    FINAL_MATCH = auto(),
    PARTIAL_MATCH = auto()


class TokenMatcher:

    def match(self, value: bytes) -> MatchType:
        return MatchType.NONE

    # TODO: Find a more efficient algorithm for string end matching
    def match_at_end(self, value: bytes) -> MatchType:
        match_length = 1
        total_length = len(value)
        while match_length <= total_length:
            partial = value[-match_length:]
            match_type = self.match(partial)
            if match_type != MatchType.NONE:
                return match_type
            match_length += 1
        return MatchType.NONE


def match_literal(literal: bytes, value: bytes) -> MatchType:
    if value == literal:
        return MatchType.FINAL_MATCH
    elif len(value) < len(literal) and literal.find(value) == 0:
        return MatchType.PARTIAL_MATCH
    else:
        return MatchType.NONE


class LiteralTokenMatcher(TokenMatcher):

    def __init__(self, value: bytes):
        self.value = value

    def match(self, value: bytes) -> MatchType:
        return match_literal(self.value, value)


class WhitespaceTokenMatcher(TokenMatcher):

    def match(self, value: bytes) -> MatchType:
        # TODO: Does this match PHP's definition of whitespace?
        if value.isspace():
            return MatchType.MATCH
        return MatchType.NONE


class OpenTagTokenMatcher(TokenMatcher):

    def match(self, value: bytes) -> MatchType:
        # TODO: Handle tag variations
        return match_literal(b'<?php', value)


DOC_COMMENT_START = b'/*'
DOC_COMMENT_END = b'*/'
COMMENT_START = b'//'
ALTERNATE_COMMENT_START = b'#'
COMMENT_END = b'\n'
CLOSING_TAG = b'?>'
POSSIBLE_COMMENT_STARTS = {
        COMMENT_START,
        ALTERNATE_COMMENT_START
    }
POSSIBLE_COMMENT_ENDS = {
        COMMENT_END,
        CLOSING_TAG
    }


class EnclosedTokenMatcher(TokenMatcher):

    def __init__(self, start: bytes, end: bytes):
        self.start = start
        self.end = end
        self.end_length = len(end)

    def match(self, value: bytes) -> MatchType:
        if value.find(self.start) == 0:
            if value.find(self.end) == \
                    len(value) - self.end_length:
                return MatchType.FINAL_MATCH
            else:
                return MatchType.PARTIAL_MATCH
        if self.start.find(value) == 0:
            return MatchType.PARTIAL_MATCH
        return MatchType.NONE


class DocCommentTokenMatcher(EnclosedTokenMatcher):

    def __init__(self):
        super().__init__(
                DOC_COMMENT_START,
                DOC_COMMENT_END
            )


class CommentTokenMatcher(TokenMatcher):

    def match(self, value: bytes) -> MatchType:
        for start in POSSIBLE_COMMENT_STARTS:
            if value.find(start) == 0:
                for end in POSSIBLE_COMMENT_ENDS:
                    if value.find(end) != -1:
                        return MatchType.NONE
                return MatchType.MATCH
            elif start.find(value) == 0:
                return MatchType.PARTIAL_MATCH
        return MatchType.NONE


VARIABLE_PREFIX = b'$'
IDENTIFIER_PATTERN = re.compile(br'^[a-zA-Z_\x80-\xff][a-zA-Z0-9_\x80-\xff]*$')


class VariableTokenMatcher(TokenMatcher):

    def match(self, value: bytes) -> MatchType:
        if value[0:1] == VARIABLE_PREFIX:
            if IDENTIFIER_PATTERN.fullmatch(value[1:]) is not None:
                return MatchType.MATCH
            if len(value) == 1:
                return MatchType.PARTIAL_MATCH
        return MatchType.NONE


class IdentifierTokenMatcher(TokenMatcher):

    def match(self, value: bytes) -> MatchType:
        if IDENTIFIER_PATTERN.fullmatch(value):
            return MatchType.MATCH
        return MatchType.NONE


STRING_QUOTES = {
    b'"',
    b"'"
}
STRING_ESCAPE = b"\\"


class StringLiteralTokenMatcher(TokenMatcher):

    def match(self, value: bytes) -> MatchType:
        quote = value[0:1]
        if quote in STRING_QUOTES:
            escaped = None
            length = len(value)
            end = length - 1
            for index in range(1, length):
                character = value[index:index+1]
                if escaped is None:
                    escaped = False
                if character == quote and not escaped:
                    if index == end:
                        return MatchType.FINAL_MATCH
                    else:
                        return MatchType.NONE
                escaped = (character == STRING_ESCAPE)
            return MatchType.MATCH
        return MatchType.NONE


INTEGER_PATTERN = re.compile(b'^[0-9]+$')


class IntegerLiteralTokenMatcher(TokenMatcher):

    def match(self, value: bytes) -> MatchType:
        # TODO: Support alternate integer syntaxes
        if INTEGER_PATTERN.match(value) is not None:
            return MatchType.MATCH
        return MatchType.NONE


class UnmatchingTokenMatcher(TokenMatcher):

    def match(self, value: bytes) -> MatchType:
        return MatchType.NONE


class CharacterTokenMatcher(TokenMatcher):

    def match(self, value: bytes) -> MatchType:
        if len(value) == 1:
            return MatchType.FINAL_MATCH
        return MatchType.NONE


def _literal(value) -> LiteralTokenMatcher:
    return LiteralTokenMatcher(value)


# TODO: Add matchers for remaining token types
class TokenType(Enum):
    # Custom matching types
    # DNUMBER   "floating-point number"
    # NAME_FULLY_QUALIFIED "fully qualified name"
    # NAME_RELATIVE "namespace-relative name"
    # NAME_QUALIFIED "namespaced name"
    # INLINE_HTML
    # ENCAPSED_AND_WHITESPACE  "string content"
    # CONSTANT_ENCAPSED_STRING "quoted string"
    # STRING_VARNAME "variable name"
    # NUM_STRING "number"
    # START_HEREDOC   "heredoc start"
    # END_HEREDOC     "heredoc end"
    OPEN_TAG = OpenTagTokenMatcher(),
    WHITESPACE = WhitespaceTokenMatcher(),
    DOC_COMMENT = DocCommentTokenMatcher(),
    COMMENT = CommentTokenMatcher(),
    VARIABLE = VariableTokenMatcher(),
    CONSTANT_ENCAPSED_STRING = StringLiteralTokenMatcher(),
    LNUMBER = IntegerLiteralTokenMatcher(),

    # Literal token types
    INCLUDE_ONCE = _literal(b'include_once'),
    INCLUDE = _literal(b'include'),
    EVAL = _literal(b'eval'),
    REQUIRE_ONCE = _literal(b'require_once'),
    REQUIRE = _literal(b'require'),
    LOGICAL_OR = _literal(b'or'),
    LOGICAL_XOR = _literal(b'xor'),
    LOGICAL_AND = _literal(b'and'),
    PRINT = _literal(b'print'),
    YIELD = _literal(b'yield'),
    YIELD_FROM = _literal(b'yield from'),
    INSTANCEOF = _literal(b'instanceof'),
    NEW = _literal(b'new'),
    CLONE = _literal(b'clone'),
    EXIT = _literal(b'exit'),
    IF = _literal(b'if'),
    ELSEIF = _literal(b'elseif'),
    ELSE = _literal(b'else'),
    ENDIF = _literal(b'endif'),
    ECHO = _literal(b'echo'),
    DO = _literal(b'do'),
    WHILE = _literal(b'while'),
    ENDWHILE = _literal(b'endwhile'),
    FOREACH = _literal(b'foreach'),
    FOR = _literal(b'for'),
    ENDFOR = _literal(b'endfor'),
    ENDFOREACH = _literal(b'endforeach'),
    DECLARE = _literal(b'declare'),
    ENDDECLARE = _literal(b'enddeclare'),
    AS = _literal(b'as'),
    SWITCH = _literal(b'switch'),
    ENDSWITCH = _literal(b'endswitch'),
    CASE = _literal(b'case'),
    DEFAULT = _literal(b'default'),
    MATCH = _literal(b'match'),
    BREAK = _literal(b'break'),
    CONTINUE = _literal(b'continue'),
    GOTO = _literal(b'goto'),
    FUNCTION = _literal(b'function'),
    FN = _literal(b'fn'),
    CONST = _literal(b'const'),
    RETURN = _literal(b'return'),
    TRY = _literal(b'try'),
    CATCH = _literal(b'catch'),
    FINALLY = _literal(b'finally'),
    THROW = _literal(b'throw'),
    USE = _literal(b'use'),
    INSTEADOF = _literal(b'insteadof'),
    GLOBAL = _literal(b'global'),
    STATIC = _literal(b'static'),
    ABSTRACT = _literal(b'abstract'),
    FINAL = _literal(b'final'),
    PRIVATE = _literal(b'private'),
    PROTECTED = _literal(b'protected'),
    PUBLIC = _literal(b'public'),
    READONLY = _literal(b'readonly'),
    VAR = _literal(b'var'),
    UNSET = _literal(b'unset'),
    ISSET = _literal(b'isset'),
    EMPTY = _literal(b'empty'),
    HALT_COMPILER = _literal(b'__halt_compiler'),
    CLASS = _literal(b'class'),
    TRAIT = _literal(b'trait'),
    INTERFACE = _literal(b'interface'),
    ENUM = _literal(b'enum'),
    EXTENDS = _literal(b'extends'),
    IMPLEMENTS = _literal(b'implements'),
    NAMESPACE = _literal(b'namespace'),
    LIST = _literal(b'list'),
    ARRAY = _literal(b'array'),
    CALLABLE = _literal(b'callable'),
    LINE = _literal(b'__LINE__'),
    FILE = _literal(b'__FILE__'),
    DIR = _literal(b'__DIR__'),
    CLASS_C = _literal(b'__CLASS__'),
    TRAIT_C = _literal(b'__TRAIT__'),
    METHOD_C = _literal(b'__METHOD__'),
    FUNC_C = _literal(b'__FUNCTION__'),
    NS_C = _literal(b'__NAMESPACE__'),
    PLUS_EQUAL = _literal(b'+='),
    MINUS_EQUAL = _literal(b'-='),
    MUL_EQUAL = _literal(b'*='),
    DIV_EQUAL = _literal(b'/='),
    CONCAT_EQUAL = _literal(b'.='),
    MOD_EQUAL = _literal(b'%='),
    AND_EQUAL = _literal(b'&='),
    OR_EQUAL = _literal(b'|='),
    XOR_EQUAL = _literal(b'^='),
    SL_EQUAL = _literal(b'<<='),
    SR_EQUAL = _literal(b'>>='),
    COALESCE_EQUAL = _literal(b'??='),
    BOOLEAN_OR = _literal(b'||'),
    BOOLEAN_AND = _literal(b'&&'),
    IS_IDENTICAL = _literal(b'==='),
    IS_NOT_IDENTICAL = _literal(b'!=='),
    IS_SMALLER_OR_EQUAL = _literal(b'<='),
    IS_GREATER_OR_EQUAL = _literal(b'>='),
    SPACESHIP = _literal(b'<=>'),
    IS_EQUAL = _literal(b'=='),
    IS_NOT_EQUAL = _literal(b'!='),
    SL = _literal(b'<<'),
    SR = _literal(b'>>'),
    INC = _literal(b'++'),
    DEC = _literal(b'--'),
    INT_CAST = _literal(b'(int)'),
    DOUBLE_CAST = _literal(b'(double)'),
    STRING_CAST = _literal(b'(string)'),
    ARRAY_CAST = _literal(b'(array)'),
    OBJECT_CAST = _literal(b'(object)'),
    BOOL_CAST = _literal(b'(bool)'),
    UNSET_CAST = _literal(b'(unset)'),
    OBJECT_OPERATOR = _literal(b'->'),
    NULLSAFE_OBJECT_OPERATOR = _literal(b'?->'),
    DOUBLE_ARROW = _literal(b'=>'),
    DOLLAR_OPEN_CURLY_BRACES = _literal(b'${'),
    CURLY_OPEN = _literal(b'{$'),
    PAAMAYIM_NEKUDOTAYIM = _literal(b'::'),
    NS_SEPARATOR = _literal(b'\\'),
    ELLIPSIS = _literal(b'...'),
    COALESCE = _literal(b'??'),
    POW = _literal(b'**'),
    POW_EQUAL = _literal(b'**='),
    ATTRIBUTE = _literal(b'#['),
    OPEN_TAG_WITH_ECHO = _literal(b'<?='),
    CLOSE_TAG = _literal(CLOSING_TAG),

    STRING = IdentifierTokenMatcher(),
    CHARACTER = CharacterTokenMatcher(),
    INLINE_HTML = UnmatchingTokenMatcher(),

    def __init__(self, matcher: TokenMatcher):
        self.matcher = matcher

    def match(self, value: bytes) -> MatchType:
        return self.matcher.match(value)

    def match_at_end(self, value: bytes) -> MatchType:
        return self.matcher.match_at_end(value)


class CharacterType(bytes, Enum):
    EQUALS = b'=',
    SEMICOLON = b';',
    OPEN_PARENTHESIS = b'(',
    CLOSE_PARENTHESIS = b')',
    COMMA = b',',
    OPEN_BRACE = b'{',
    CLOSE_BRACE = b'}'


class Token:

    def __init__(self, type: TokenType, value: bytes):
        self.type = type
        self.value = value

    def is_character(
                self,
                type: Optional[Union[CharacterType, Set[CharacterType]]] = None
            ) -> bool:
        return self.type is TokenType.CHARACTER \
            and (
                type is None
                or (isinstance(type, CharacterType) and self.value == type)
                or self.value in type
            )

    def is_semicolon(self) -> bool:
        return self.is_character(CharacterType.SEMICOLON)

    def is_opening_parenthesis(self) -> bool:
        return self.is_character(CharacterType.OPEN_PARENTHESIS)

    def is_closing_parenthesis(self) -> bool:
        return self.is_character(CharacterType.CLOSE_PARENTHESIS)

    def is_comma(self) -> bool:
        return self.is_character(CharacterType.COMMA)

    def __repr__(self) -> str:
        return self.type.name + ' (' + bytes_to_str(self.value) + ')'

    def __str__(self) -> str:
        return repr(self)


class Lexer:

    def __init__(self, stream: BinaryIO, chunk_size: int = 4096):
        self.chunks = deque()
        self.chunk_size = chunk_size
        self.chunk_offset = 0
        self.read = 0
        self.stream = stream
        self.offset = 0
        self.position = 0
        self.inside_tag = False

    def _read_chunk(self) -> bool:
        chunk = self.stream.read(self.chunk_size)
        length = len(chunk)
        if length == 0:
            return False
        self.read += length
        self.chunks.append(chunk)
        return True

    def step(self) -> bool:
        self.position += 1
        if self.position > self.read:
            if not self._read_chunk():
                return False
        return True

    def get_current(self) -> bytes:
        components = []
        remaining = self.position - self.offset
        remaining_offset = self.offset
        for chunk in self.chunks:
            chunk_length = len(chunk)
            if chunk_length <= remaining_offset:
                remaining_offset -= chunk_length
                continue
            chunk = chunk[remaining_offset:remaining_offset + remaining]
            components.append(chunk)
            chunk_length = len(chunk)
            if chunk_length >= remaining:
                break
            remaining -= chunk_length
        return b''.join(components)

    def step_backwards(self) -> None:
        self.position -= 1

    def reset(self) -> None:
        self.position = self.offset

    def consume_token(self, token_type: TokenType) -> Token:
        value = self.get_current()
        self.offset = self.position
        return Token(token_type, value)

    def extract_php_token(self, types=TokenType) -> Optional[Token]:
        while self.step():
            position = self.position
            for type in types:
                current = self.get_current()
                potential = False
                while True:
                    match_type = type.match(current)
                    if match_type == MatchType.FINAL_MATCH:
                        position = self.position
                        if TokenType.STRING.match(current) == MatchType.MATCH:
                            string = self.extract_php_token({TokenType.STRING})
                            if string is not None:
                                return string
                        self.position = position
                        return self.consume_token(type)
                    elif match_type == MatchType.MATCH:
                        potential = True
                    elif match_type == MatchType.NONE:
                        if potential:
                            self.step_backwards()
                            return self.consume_token(type)
                        self.position = position
                        break
                    if not self.step():
                        if potential:
                            return self.consume_token(type)
                        self.position = position
                        break
                    current = self.get_current()
        current = self.get_current()
        if current is not None and len(current) > 0 and types is TokenType:
            raise LexingException(
                    f'Input does not match any known token: "{current}"'
                )
        return None

    def extract_inline_html_or_open_tag(self) -> Optional[Token]:
        partial_start = None
        while self.step():
            current = self.get_current()
            match_type = TokenType.OPEN_TAG.match_at_end(current)
            if match_type == MatchType.PARTIAL_MATCH and partial_start is None:
                partial_start = self.position
            elif match_type == MatchType.FINAL_MATCH:
                if partial_start is None or partial_start > 0:
                    return self.consume_token(TokenType.OPEN_TAG)
                else:
                    self.offset = partial_start
                    return self.consume_token(TokenType.INLINE_HTML)
            else:
                partial_start = None
        current = self.get_current()
        if current is not None and len(current) > 0:
            return self.consume_token(TokenType.INLINE_HTML)
        return None

    def get_next_token(self) -> Optional[Token]:
        if self.inside_tag:
            token = self.extract_php_token()
            if token is not None and token.type == TokenType.CLOSE_TAG:
                self.inside_tag = False
        else:
            token = self.extract_inline_html_or_open_tag()
            if token is not None and token.type == TokenType.OPEN_TAG:
                self.inside_tag = True
        return token


def lex(stream: BinaryIO) -> Generator[Token, None, None]:
    lexer = Lexer(stream)
    while (token := lexer.get_next_token()) is not None:
        yield token



import os
import pickle
import base64
import time
import shutil
from typing import Any, Callable, Optional, Set, Iterable

from .io import FileLock, LockType
from .serialization import limited_deserialize
from ..logging import log


DURATION_ONE_DAY = 86400


class CacheException(Exception):
    pass


class NoCachedValueException(CacheException):
    pass


class InvalidCachedValueException(CacheException):
    pass


CacheFilter = Callable[[Any], Any]


class Cache:

    def __init__(self):
        self.filters = []

    def _serialize_value(self, value: Any) -> Any:
        return value

    def _deserialize_value(self, value: Any) -> Any:
        return value

    def _save(self, key: str, value: Any) -> None:
        raise NotImplementedError('Saving is not implemented for this cache')

    def _load(self, key: str, max_age: Optional[int]) -> Any:
        raise NotImplementedError('Loading is not implemented for this cache')

    def put(self, key: str, value) -> None:
        self._save(key, self._serialize_value(value))

    def get(
                self,
                key: str,
                max_age: Optional[int] = None,
                additional_filters: Optional[Iterable[CacheFilter]] = None
            ) -> Any:
        return self.filter_value(
                self._deserialize_value(self._load(key, max_age)),
                additional_filters
            )

    def remove(self, key: str) -> None:
        raise NotImplementedError('Removing is not implemented for this cache')

    def purge(self) -> None:
        pass

    def add_filter(self, filter: CacheFilter) -> None:
        self.filters.append(filter)

    def filter_value(
                self,
                value: Any,
                additional_filters: Optional[Iterable[CacheFilter]] = None
            ) -> Any:
        for filter in self.filters:
            value = filter(value)
        if additional_filters is not None:
            for filter in additional_filters:
                value = filter(value)
        return value


class RuntimeCache(Cache):

    def __init__(self):
        super().__init__()
        self.purge()

    def _save(self, key: str, value: Any) -> None:
        self.items[key] = value

    def _load(self, key: str, max_age: Optional[int]) -> Any:
        if key in self.items:
            return self.items[key]
        raise NoCachedValueException()

    def remove(self, key: str) -> None:
        try:
            del self.items[key]
        except KeyError:
            # The item already does not exist
            pass

    def purge(self) -> None:
        self.items = {}


class CacheDirectory(Cache):

    def __init__(self, path: bytes, allowed: Optional[Set[str]] = None):
        super().__init__()
        self.path = path
        self.allowed = allowed
        self._initialize_directory()

    def _initialize_directory(self) -> None:
        try:
            os.makedirs(self.path, mode=0o700, exist_ok=True)
        except OSError as e:
            raise CacheException(
                    f'Failed to initialize cache directory at {self.path}'
                ) from e

    def _serialize_value(self, value: Any) -> Any:
        return pickle.dumps(value)

    def _deserialize_value(self, value: Any) -> Any:
        return limited_deserialize(value, self.allowed)

    def _get_path(self, key: str) -> bytes:
        return os.path.join(
                self.path,
                os.fsencode(base64.b16encode(key.encode('utf-8')))
            )

    def _save(self, key: str, value: Any) -> None:
        path = self._get_path(key)
        with open(path, 'wb') as file:
            with FileLock(file, LockType.EXCLUSIVE):
                file.write(value)

    def _is_valid(self, path: str, max_age: Optional[int]) -> bool:
        if max_age is None:
            return True
        modified_timestamp = os.path.getmtime(path)
        current_timestamp = time.time()
        return current_timestamp - modified_timestamp < max_age

    def _load(self, key: str, max_age: Optional[int]) -> Any:
        path = self._get_path(key)
        try:
            with open(path, 'rb') as file:
                with FileLock(file, LockType.SHARED):
                    if not self._is_valid(path, max_age):
                        os.remove(path)
                        raise NoCachedValueException()
                    value = file.read()
            return value
        except OSError as e:
            if not isinstance(e, FileNotFoundError):
                log.warning(
                        'Unexpected error occurred while reading from cache: '
                        + str(e)
                    )
            raise NoCachedValueException() from e

    def remove(self, key: str) -> None:
        path = self._get_path(key)
        with open(path, 'wb') as file:
            with FileLock(file, LockType.EXCLUSIVE):
                os.remove(path)

    def purge(self) -> None:
        try:
            shutil.rmtree(self.path)
        except BaseException as e:  # noqa: B036
            raise CacheException('Failed to delete cache directory') from e
        self._initialize_directory()


class Cacheable:

    def __init__(
                self,
                key: str,
                initializer: Callable[[], Any],
                max_age: Optional[int] = None,
                filters: Optional[Iterable[CacheFilter]] = None
            ):
        self.key = key
        self._initializer = initializer
        self.max_age = max_age
        self.filters = filters

    def _initialize_value(self) -> Any:
        return self._initializer()

    def get(self, cache: Cache) -> Any:
        try:
            value = cache.get(self.key, self.max_age, self.filters)
        except (
                NoCachedValueException,
                InvalidCachedValueException
                ):
            value = self._initialize_value()
            self.set(cache, value)
        return value

    def set(self, cache: Cache, value: Any) -> None:
        cache.put(self.key, value)

    def delete(self, cache: Cache) -> Any:
        cache.remove(self.key)



from wordfence.wordpress.database import DEFAULT_HOST, DEFAULT_PORT, \
    DEFAULT_USER, DEFAULT_PREFIX, DEFAULT_COLLATION

from ..subcommands import SubcommandDefinition, UsageExample
from ..config.typing import ConfigDefinitions

from .reporting import DATABASE_SCAN_REPORT_CONFIG_OPTIONS

config_definitions: ConfigDefinitions = {
    "host": {
        "short_name": "H",
        "description": "The database hostname",
        "context": "CLI",
        "argument_type": "OPTION",
        "default": DEFAULT_HOST,
        "category": "Database Connectivity"
    },
    "port": {
        "short_name": "P",
        "description": "The database port",
        "context": "CLI",
        "argument_type": "OPTION",
        "default": DEFAULT_PORT,
        "meta": {
            "value_type": int
        },
        "category": "Database Connectivity"
    },
    "user": {
        "short_name": "u",
        "description": "The database user",
        "context": "CLI",
        "argument_type": "OPTION",
        "default": DEFAULT_USER,
        "category": "Database Connectivity"
    },
    "password": {
        "description": "The database password (this option is insecure)",
        "context": "CLI",
        "argument_type": "OPTION",
        "default": None,
        "category": "Database Connectivity"
    },
    "prompt-for-password": {
        "short_name": "p",
        "description": "Prompt for a database password",
        "context": "CLI",
        "argument_type": "FLAG",
        "default": False,
        "category": "Database Connectivity"
    },
    "password-env": {
        "description": "The environment variable name to check for a password",
        "context": "ALL",
        "argument_type": "OPTION",
        "default": "WFCLI_DB_PASSWORD",
        "category": "Database Connectivity"
    },
    "prefix": {
        "short_name": "x",
        "description": "The WordPress database prefix",
        "context": "CLI",
        "argument_type": "OPTION",
        "default": DEFAULT_PREFIX,
        "category": "Database Connectivity"
    },
    "database-name": {
        "short_name": "D",
        "description": "The MySQL database name",
        "context": "CLI",
        "argument_type": "OPTION",
        "default": None,
        "category": "Database Connectivity"
    },
    "collation": {
        "short_name": "C",
        "description": "The collation to use when connecting to MySQL",
        "context": "CLI",
        "argument_type": "OPTION",
        "default": DEFAULT_COLLATION
    },
    "read-stdin": {
        "description": "Read paths from stdin. If not specified, paths will "
                       "automatically be read from stdin when input is not "
                       "from a TTY.",
        "context": "ALL",
        "argument_type": "OPTIONAL_FLAG",
        "default": None
    },
    "path-separator": {
        "short_name": "s",
        "description": "Separator used to delimit paths when reading from "
                       "stdin. Defaults to the null byte.",
        "context": "ALL",
        "argument_type": "OPTION",
        "default": "AA==",
        "default_type": "base64"
    },
    **DATABASE_SCAN_REPORT_CONFIG_OPTIONS,
    "require-database": {
        "description": "When enabled, invoking the db-scan command without "
                       "specifying at least one database will trigger an "
                       "error. This is the default behavior when running in "
                       "a terminal.",
        "context": "CLI",
        "argument_type": "OPTIONAL_FLAG",
        "default": None
    },
    "locate-sites": {
        "short_name": "S",
        "description": (
                "Automatically locate WordPress config files to extract "
                "database connection details"
            ),
        "context": "CLI",
        "argument_type": "FLAG",
        "default": None,
        "category": "Site Location"
    },
    "allow-nested": {
        "description": "Allow WordPress installations nested below other "
                       "installations to be identified as targets for "
                       "database scanning",
        "context": "ALL",
        "argument_type": "FLAG",
        "default": True,
        "category": "Site Location"
    },
    "allow-io-errors": {
        "description": "Allow scanning to continue even if an IO error occurs "
                       "while locating WordPress sites. Sites that cannot "
                       "be identified due to IO errors will be excluded from "
                       "scanning. This is the default behavior.",
        "context": "ALL",
        "argument_type": "FLAG",
        "default": True,
        "category": "Site Location"
    },
    "use-remote-rules": {
        "description": "If enabled, scanning rules will be pulled from "
                       "the Wordfence API. This is the default behavior.",
        "context": "ALL",
        "argument_type": "FLAG",
        "default": True
    },
    "rules-file": {
        "short_name": "R",
        "description": "Path to a JSON file containing scanning rules",
        "context": "ALL",
        "argument_type": "OPTION_REPEATABLE",
        "meta": {
            "accepts_file": True
        }
    },
    "exclude-rules": {
        "short_name": "e",
        "description": "Specify rule IDs to ignore when scanning. May be "
                       "comma-delimited and/or repeated.",
        "context": "ALL",
        "argument_type": "OPTION_REPEATABLE",
        "default": None,
        "meta": {
            "separator": ",",
            "value_type": int
        }
    },
    "include-rules": {
        "short_name": "i",
        "description": "Specify rule IDs to include when scanning. May be "
                       "comma-delimited and/or repeated.",
        "context": "ALL",
        "argument_type": "OPTION_REPEATABLE",
        "default": None,
        "meta": {
            "separator": ",",
            "value_type": int
        }
    }
}

examples = [
    UsageExample(
        'Scan the WordPress database at db.example.com',
        'wordfence db-scan -h db.example.com -p wordpress'
    )
]

definition = SubcommandDefinition(
    name='db-scan',
    usage=[
        '[OPTIONS] [DATABASE_CONFIG_PATH]...',
        '[OPTIONS] -S [WORDPRESS_INSTALLATION_PATH]...'
    ],
    description='Scan for malicious content in WordPress databases',
    config_definitions=config_definitions,
    config_section='DB_SCAN',
    cacheable_types={
            'wordfence.intel.database_rules.DatabaseRuleSet',
            'wordfence.intel.database_rules.DatabaseRule'
        },
    examples=examples,
    accepts_directories=True
)



from enum import Enum
from platform import machine
from typing import Set


class UnknownPlatform(Exception):
    pass


class Platform(Enum):

    AMD64 = ('amd64', {'amd64', 'x86_64'})
    ARM64 = ('arm64', {'arm64', 'aarch64'})

    def __init__(self, key: str, machine_names: Set[str]):
        self.key = key
        self.machine_names = machine_names

    @classmethod
    def detect(cls):
        machine_name = machine()
        for platform in cls:
            if machine_name in platform.machine_names:
                return platform
        raise UnknownPlatform()



from ..subcommands import SubcommandDefinition, UsageExample
from ..config.typing import ConfigDefinitions
from ...api.intelligence import VulnerabilityFeedVariant
from .reporting import VULN_SCAN_REPORT_CONFIG_OPTIONS

config_definitions: ConfigDefinitions = {
    "read-stdin": {
        "description": "Read WordPress base paths from stdin. If not specified"
                       ", paths will automatically be read from stdin when "
                       "input is not from a TTY.",
        "context": "ALL",
        "argument_type": "OPTIONAL_FLAG",
        "default": None
    },
    "path-separator": {
        "short_name": "s",
        "description": "Separator used to delimit paths when reading from "
                       "stdin. Defaults to the null byte.",
        "context": "ALL",
        "argument_type": "OPTION",
        "default": "AA==",
        "default_type": "base64"
    },
    "wordpress-path": {
        "short_name": "w",
        "description": "Path to the root of a WordPress installation to scan"
                       " for core vulnerabilities.",
        "context": "CLI",
        "argument_type": "OPTION_REPEATABLE",
        "default": [],
        "meta": {
            "separator": ",",
            "accepts_directory": True
        }
    },
    "plugin-directory": {
        "short_name": "p",
        "description": "Path to a directory containing WordPress plugins to"
                       " scan for vulnerabilities.",
        "context": "CLI",
        "argument_type": "OPTION_REPEATABLE",
        "default": [],
        "meta": {
            "separator": ",",
            "accepts_directory": True
        }
    },
    "theme-directory": {
        "short_name": "t",
        "description": "Path to a directory containing WordPress themes to"
                       " scan for vulnerabilities.",
        "context": "CLI",
        "argument_type": "OPTION_REPEATABLE",
        "default": [],
        "meta": {
            "separator": ",",
            "accepts_directory": True
        }
    },
    "relative-content-path": {
        "short_name": "C",
        "description": "Alternate path of the wp-content directory relative "
                       "to the WordPress root.",
        "context": "ALL",
        "argument_type": "OPTION_REPEATABLE",
        "default": [],
        "meta": {
            "separator": ","
        }
    },
    "relative-plugins-path": {
        "short_name": "P",
        "description": "Alternate path of the wp-content/plugins directory "
                       "relative to the WordPress root.",
        "context": "ALL",
        "argument_type": "OPTION_REPEATABLE",
        "default": [],
        "meta": {
            "separator": ","
        }
    },
    "relative-mu-plugins-path": {
        "short_name": "M",
        "description": "Alternate path of the wp-content/mu-plugins directory "
                       "relative to the WordPress root.",
        "context": "ALL",
        "argument_type": "OPTION_REPEATABLE",
        "default": [],
        "meta": {
            "separator": ","
        }
    },
    **VULN_SCAN_REPORT_CONFIG_OPTIONS,
    "exclude-vulnerability": {
        "short_name": "e",
        "description": "Vulnerability UUIDs or CVE IDs to exclude from scan "
                       "results.",
        "context": "ALL",
        "argument_type": "OPTION_REPEATABLE",
        "default": [],
        "meta": {
            "separator": ","
        }
    },
    "include-vulnerability": {
        "short_name": "i",
        "description": "Vulnerabilitiy UUIDs or CVE IDs to include in scan "
                       "results.",
        "context": "ALL",
        "argument_type": "OPTION_REPEATABLE",
        "default": [],
        "meta": {
            "separator": ","
        }
    },
    "informational": {
        "short_name": "I",
        "description": "Include informational vulnerability records in "
                       "results.",
        "context": "ALL",
        "argument_type": "FLAG",
        "default": False
    },
    "feed": {
        "short_name": "f",
        "description": "The feed to use for vulnerability information. "
                       "The production feed provides all available "
                       "information fields. The scanner feed contains "
                       "only the minimum fields necessary to conduct a scan "
                       "and may be a better choice when detailed "
                       "vulnerability information is not needed.",
        "context": "ALL",
        "argument_type": "OPTION",
        "default": VulnerabilityFeedVariant.PRODUCTION.path,
        "meta": {
            "valid_options": [
                    variant.path for variant in VulnerabilityFeedVariant
                ]
        }
    },
    "require-path": {
        "description": "When enabled, an error will be issued if at least one "
                       "path to scan is not specified. This is the default "
                       "behavior when running in a terminal.",
        "context": "CLI",
        "argument_type": "OPTIONAL_FLAG",
        "default": None
    },
    "allow-nested": {
        "description": "When enabled (the default), WordPress installations "
                       "nested below other installations will also be "
                       "scanned for vulnerabilities.",
        "context": "ALL",
        "argument_type": "FLAG",
        "default": True
    },
    "allow-io-errors": {
        "description": "Allow scanning to continue if IO errors are "
                       "encountered. Sites that cannot be processed "
                       "due to IO errors will be skipped and a warning will "
                       "be logged. This is the default behavior.",
        "context": "ALL",
        "argument_type": "FLAG",
        "default": True
    }
}

cacheable_types = {
    'wordfence.intel.vulnerabilities.VulnerabilityIndex',
    'wordfence.intel.vulnerabilities.ScannerVulnerability',
    'wordfence.intel.vulnerabilities.ProductionVulnerability',
    'wordfence.intel.vulnerabilities.Software',
    'wordfence.intel.vulnerabilities.ProductionSoftware',
    'wordfence.intel.vulnerabilities.SoftwareType',
    'wordfence.intel.vulnerabilities.VersionRange',
    'wordfence.intel.vulnerabilities.CopyrightInformation',
    'wordfence.intel.vulnerabilities.Copyright',
    'wordfence.intel.vulnerabilities.Cwe',
    'wordfence.intel.vulnerabilities.Cvss'
}

examples = [
    UsageExample(
        'Scan the WordPress installation at /var/www/html for vulnerabilities',
        'wordfence vuln-scan /var/www/html'
    ),
    UsageExample(
        'Generate a CSV file containing vulnerabilities found after scanning '
        '/var/www/html',
        'wordfence vuln-scan --output-format csv --output-path '
        '/tmp/wfcli-results.csv --output-columns link /var/www/html'
    )
]

definition = SubcommandDefinition(
    name='vuln-scan',
    usage='[OPTIONS] [WORDPRESS_PATH]...',
    description='Scan WordPress installations for vulnerable software',
    config_definitions=config_definitions,
    config_section='VULN_SCAN',
    cacheable_types=cacheable_types,
    examples=examples,
    accepts_directories=True
)



from ...util import caching
from ...logging import log
from ..subcommands import Subcommand
from ..mailing_lists import EMAIL_SIGNUP_MESSAGE


class TermsSubcommand(Subcommand):

    def fetch_terms(self) -> str:
        client = self.context.create_noc1_client(use_hooks=False)
        return client.get_terms()

    def get_terms(self) -> str:
        cacheable = caching.Cacheable(
                'terms',
                self.fetch_terms,
                caching.DURATION_ONE_DAY
            )
        return cacheable.get(self.cache)

    def invoke(self) -> int:
        terms = self.get_terms()
        print(terms)
        log.info(EMAIL_SIGNUP_MESSAGE)
        return 0


factory = TermsSubcommand



import os

from typing import List, Dict, Tuple
from dataclasses import dataclass

from ..helper import Helper
from .cli_parser import CliCanonicalValueExtractor, get_cli_values
from .config_items import ConfigItemDefinition, \
    CanonicalValueExtractorInterface, not_set_token
from .ini_parser import load_ini, get_ini_value_extractor, \
        get_default_ini_value_extractor
from ..subcommands import SubcommandDefinition
from .base_config_definitions import config_map as base_config_map
from .config import Config


value_extractors: List = []


class RenamedSubcommandException(Exception):

    def __init__(self, old: str, new: str):
        self.old = old
        self.new = new


def create_config_object(
            subcommand: str,
            definitions: Dict[str, ConfigItemDefinition],
            trailing_arguments: List[str],
            parser,
            *ordered_sources
        ):
    if len(ordered_sources) < 1:
        raise ValueError("At least one configuration source must be passed in")
    target = Config(definitions, parser, subcommand)
    for source in ordered_sources:
        source_extractors: List[CanonicalValueExtractorInterface] = []
        for extractor in value_extractors:
            if extractor.is_valid_source(source):
                source_extractors.append(extractor)
        if len(source_extractors) == 0:
            raise Exception(
                    'No compatible extractor found for provided config source'
                )
        # extract all values from the source and
        # conditionally update the config
        for item_definition in definitions.values():
            new_value = not_set_token
            for extractor in source_extractors:
                new_value = extractor.get_canonical_value(
                        item_definition,
                        source
                    )
                if new_value is not not_set_token:
                    break

            # later values always replace previous values
            if new_value is not not_set_token:
                setattr(target, item_definition.property_name, new_value)
                target.sources[item_definition.property_name] = \
                    extractor.get_context()
                try:
                    target.defaulted_options.remove(
                            item_definition.property_name
                        )
                except KeyError:
                    pass  # Ignore options that weren't previously defaulted
            elif not hasattr(target, item_definition.property_name):
                default = item_definition.default
                if item_definition.has_separator() and \
                        isinstance(default, str):
                    default = default.split(item_definition.meta.separator)
                setattr(target, item_definition.property_name,
                        default)
                target.defaulted_options.add(item_definition.property_name)
    target.trailing_arguments = trailing_arguments
    return target


def _get_renamed_subcommand(
            subcommand: str,
            definitions: Dict[str, SubcommandDefinition]
        ) -> str:
    for definition in definitions.values():
        if subcommand in definition.previous_names:
            return definition.name
    raise KeyError(
            f'Subcommand {subcommand} does not appear to have been renamed'
        )


def resolve_config_map(subcommand_definition: SubcommandDefinition):
    return {
            **base_config_map,
            **subcommand_definition.get_config_map()
        }


@dataclass
class GlobalConfig:
    debug: bool = False


def load_config(
            subcommand_definitions: Dict[str, SubcommandDefinition],
            helper: Helper,
            subcommand: str = None,
            global_config: GlobalConfig = None
        ) -> Tuple[Config, SubcommandDefinition]:
    cli_values, trailing_arguments, parser = get_cli_values(
            subcommand_definitions,
            helper
        )

    if global_config is not None:
        global_config.debug = False if cli_values.debug is not_set_token \
            else cli_values.debug

    if subcommand is None:
        subcommand = cli_values.subcommand

    if subcommand:
        try:
            subcommand_definition = subcommand_definitions[subcommand]
        except KeyError:
            raise RenamedSubcommandException(
                    subcommand,
                    _get_renamed_subcommand(
                        subcommand,
                        subcommand_definitions
                    )
                )
        config_map = resolve_config_map(subcommand_definition)

    else:
        subcommand_definition = None
        config_map = base_config_map

    ini_values, ini_path = load_ini(cli_values, subcommand_definition)

    trailing_arguments_are_paths = False
    if subcommand_definition is not None:
        value_extractors.append(get_ini_value_extractor(subcommand_definition))
        trailing_arguments_are_paths = subcommand_definition.accepts_paths()
    value_extractors.append(get_default_ini_value_extractor())
    value_extractors.append(CliCanonicalValueExtractor())

    instance = create_config_object(
            subcommand,
            config_map,
            trailing_arguments,
            parser,
            ini_values,
            cli_values
        )

    if trailing_arguments_are_paths:
        instance.trailing_arguments = [
                os.fsencode(path) for path in instance.trailing_arguments
            ]

    if global_config is not None:
        global_config.debug = instance.debug

    instance.ini_path = ini_path
    return instance, subcommand_definition



import os

from typing import Optional

from ..util.io import iterate_files, resolve_path
from ..logging import log
from ..api import noc1
from ..api.exceptions import ApiException
from .identifier import FileIdentifier, FileType, FileIdentity, \
    KnownFileIdentity, GroupIdentity


class RemediationSource:

    def get_correct_content(
                self,
                identity: KnownFileIdentity
            ) -> Optional[bytes]:
        pass


class Noc1RemediationSource(RemediationSource):

    def __init__(self, client: noc1.Client):
        self.client = client

    def get_correct_content(
                self,
                identity: KnownFileIdentity
            ) -> Optional[bytes]:
        try:
            type = identity.type.value
            path = os.fsdecode(identity.local_path)
            if identity.extension is None:
                return self.client.get_wp_file_content(
                        type,
                        path,
                        identity.site.get_version(),
                    )
            else:
                return self.client.get_wp_file_content(
                        type,
                        path,
                        identity.site.get_version(),
                        identity.extension.get_name(),
                        identity.extension.version
                    )
        except ApiException as e:
            log.warning(
                    f'Unable to fetch correct content for {identity} - {e}'
                )
            return None


class RemediationResult:

    def __init__(
                self,
                path: bytes,
                identity: FileIdentity,
                known: bool = False,
                remediated: bool = False,
                target_path: Optional[bytes] = None
            ):
        self.path = path
        self.identity = identity
        self.known = known
        self.remediated = remediated
        self.target_path = target_path if target_path is not None else path

    def __bool__(self) -> bool:
        return self.remediated


class Remediator:

    def __init__(self, source: RemediationSource):
        self.identifier = FileIdentifier()
        self.source = source
        self.input_count = 0

    def get_correct_content(self, identity: KnownFileIdentity) -> bytes:
        return self.source.get_correct_content(identity)

    def remediate_file(
                self,
                path: bytes,
                target_path: Optional[bytes] = None
            ) -> RemediationResult:
        identity = self.identifier.identify(path)
        result = RemediationResult(path, identity, target_path=target_path)
        if identity.type is FileType.UNKNOWN:
            log.warning(f'Unable to identify {path}')
            return result
        log.debug('Identified ' + os.fsdecode(path) + f' as {identity}')
        correct_content = self.get_correct_content(identity)
        if correct_content is None:
            log.warning(
                    f'Unable to determine correct content for {identity}, '
                    'skipping remediation...'
                )
            return result
        result.known = True
        try:
            log.debug('Overwriting ' + os.fsdecode(path) + '...')
            with open(path, 'wb') as file:
                file.write(correct_content)
            result.remediated = True
        except OSError as error:
            log.error(
                    f'An error occurred while attempting to remediate {path}: '
                    + str(error)
                )
        return result

    def handle_symlink_loop(self, path: str) -> None:
        log.warning(f'Symlink loop detected at {path}')

    def remediate(self, path: bytes) -> RemediationResult:
        self.input_count += 1
        path = resolve_path(path)
        if os.path.isdir(path):
            file_found = False
            for file in iterate_files(
                        path,
                        loop_callback=self.handle_symlink_loop
                    ):
                yield self.remediate_file(resolve_path(file), path)
                file_found = True
            if not file_found:
                yield RemediationResult(
                        path,
                        GroupIdentity(FileType.UNKNOWN, path)
                    )
        else:
            yield self.remediate_file(path)



import logging

from ..util.terminal import Color, escape, RESET


class ConfigurableFormatter(logging.Formatter):

    def __init__(self, colored: bool = False, prefixed: bool = False):
        super().__init__()
        self.colored = colored
        self.prefixed = prefixed
        self.reset = RESET if colored else ''

    def get_style(self, level) -> str:
        if not self.colored:
            return ''
        if level >= logging.ERROR:
            return escape(color=Color.RED)
        if level >= logging.WARNING:
            return escape(color=Color.YELLOW)
        if level <= logging.DEBUG:
            return escape(color=Color.WHITE)
        return escape(color=Color.GREEN)

    def get_prefix(self, level) -> str:
        if not self.prefixed or not level:
            return ''
        return f'{level}: '

    def format(self, record) -> str:
        style = self.get_style(record.levelno)
        prefix = self.get_prefix(record.levelname)
        message = super().format(record)
        return f'{style}{prefix}{message}{self.reset}'



from ..subcommands import SubcommandDefinition
from ..config.typing import ConfigDefinitions

config_definitions: ConfigDefinitions = {}

cacheable_types = set()

definition = SubcommandDefinition(
    name='version',
    usage='',
    description='Display the version of Wordfence CLI',
    config_definitions=config_definitions,
    config_section='DEFAULT',
    cacheable_types=cacheable_types,
    requires_config=False
)



from hashlib import sha256
from typing import Optional

from ..api.licensing import License, LicenseSpecific
from ..util.serialization import limited_deserialize


class CommonString:

    def __init__(self, string: str, signature_ids: list = None):
        self.string = string
        if signature_ids is None:
            signature_ids = []
        self.signature_ids = signature_ids


class Signature:

    def __init__(
                self,
                identifier: int,
                rule: str,
                name: str,
                description: str,
                common_strings: list = None
            ):
        self.identifier = identifier
        self.rule = rule
        self.name = name
        self.description = description
        self.common_strings = common_strings \
            if common_strings is not None \
            else []

    def get_common_string_count(self) -> int:
        return len(self.common_strings)

    def has_common_strings(self) -> bool:
        return self.get_common_string_count() > 0


class SignatureSet(LicenseSpecific):

    def __init__(
                self,
                common_strings: list,
                signatures: dict,
                license: Optional[License] = None
            ):
        super().__init__(license)
        self.common_strings = common_strings
        self.signatures = signatures

    def remove_signature(self, identifier: int) -> bool:
        if identifier not in self.signatures:
            return False
        signature = self.signatures[identifier]
        for index in signature.common_strings:
            self.common_strings[index].signature_ids.remove(identifier)
        del self.signatures[identifier]
        return True

    def get_signature(self, identifier: int) -> Signature:
        if identifier in self.signatures:
            return self.signatures[identifier]
        raise ValueError(f'Invalid signature identifier: {identifier}')

    def has_signature(self, identifier: int) -> bool:
        try:
            self.get_signature(identifier)
            return True
        except ValueError:
            return False

    def get_hash(self) -> str:
        hash = sha256()
        delimiter = ';'
        for signature in self.signatures.values():
            hash.update(delimiter.join([
                    str(signature.identifier),
                    signature.rule,
                    delimiter.join(
                            [self.common_strings[index].string for index
                                in signature.common_strings]
                        )
                ]).encode('utf-8'))
        return hash.digest()


class PrecompiledSignatureSet(LicenseSpecific):

    VERSION = 1

    def __init__(
                self,
                signature_set: SignatureSet,
                data: bytes,
                signature_hash: Optional[bytes] = None,
                license: Optional[License] = None
            ):
        super().__init__(license)
        self.signature_set = signature_set
        self.signature_hash = (
                signature_set if isinstance(signature_set, bytes)
                else signature_set.get_hash()
            )
        self.data = data
        self.version = self.VERSION

    def is_supported_version(self) -> bool:
        return hasattr(self, 'version') and self.version == self.VERSION

    def assign_license(self, license: Optional[License]):
        super().assign_license(license)
        self.signature_set.assign_license(license)


def deserialize_precompiled_signature_set(data):
    signature_set = limited_deserialize(
            data,
            {
                'wordfence.api.licensing.License',
                'wordfence.intel.signatures.PrecompiledSignatureSet',
                'wordfence.intel.signatures.SignatureSet',
                'wordfence.intel.signatures.Signature',
                'wordfence.intel.signatures.CommonString'
            },
            PrecompiledSignatureSet
        )
    return signature_set



import json
from typing import Any
from base64 import b64encode


UNFILTERED_TYPES = {
        bool,
        int,
        float,
        str
    }


def encode_invalid_data(data) -> Any:
    for unfiltered_type in UNFILTERED_TYPES:
        if isinstance(data, unfiltered_type):
            return data
    if isinstance(data, dict):
        filtered = {}
        for key, value in data.items():
            filtered[encode_invalid_data(key)] = encode_invalid_data(value)
        return filtered
    elif isinstance(data, list):
        filtered = []
        for value in data:
            filtered.append(encode_invalid_data(value))
        return filtered
    elif isinstance(data, bytes):
        return b64encode(data).decode('utf-8')
    else:
        try:
            json.dumps(data)
        except Exception:
            return None


# Encode any data that cannot be represented as valid JSON
# prior to attempting to encode data as JSON
def safe_json_encode(data) -> str:
    return json.dumps(encode_invalid_data(data))



from wordfence.wordpress.database import WordpressDatabase, \
    WordpressDatabaseServer, DEFAULT_PORT, DEFAULT_COLLATION, DEFAULT_PREFIX
from wordfence.wordpress.site import WordpressLocator, \
    WordpressSite
from wordfence.wordpress.exceptions import WordpressException
from wordfence.intel.database_rules import DatabaseRuleSet, load_database_rules
from wordfence.databasescanning.scanner import DatabaseScanner
from wordfence.util.validation import ListValidator, DictionaryValidator, \
    OptionalValueValidator
from wordfence.util import caching
from getpass import getpass
from typing import Optional, List, Generator
import os
import json

from ...logging import log
from ..subcommands import Subcommand
from ..io import IoManager
from ..exceptions import ConfigurationException
from ..config import not_set_token

from .reporting import DatabaseScanReportManager


class DbScanSubcommand(Subcommand):

    def _resolve_password(self) -> Optional[str]:
        if self.config.password is not None:
            log.warning(
                    'Providing passwords via command line parameters is '
                    'insecure as they can be exposed to other users'
                )
            return self.config.password
        elif self.config.prompt_for_password:
            return getpass()
        return os.environ.get(self.config.password_env)

    def _get_base_database(self) -> Optional[WordpressDatabase]:
        name = self.config.database_name
        if name is None:
            return None
        server = WordpressDatabaseServer(
                host=self.config.host,
                port=self.config.port,
                user=self.config.user,
                password=self._resolve_password()
            )
        return WordpressDatabase(
                name=name,
                server=server,
                prefix=self.config.prefix,
                collation=self.config.collation
            )

    def _get_search_paths(
                self,
                io_manager: IoManager,
                include_current: bool = False
            ) -> Generator[bytes, None, None]:
        if len(self.config.trailing_arguments):
            yield from self.config.trailing_arguments
        elif include_current and not io_manager.should_read_stdin():
            yield os.fsencode(os.getcwd())
        if io_manager.should_read_stdin():
            for path in io_manager.get_input_reader().read_all_entries():
                yield path

    def _locate_site_databases(
                self,
                io_manager: IoManager
            ) -> Generator[WordpressDatabase, None, None]:
        for path in self._get_search_paths(io_manager, include_current=True):
            locator = WordpressLocator(
                    path=path,
                    allow_nested=self.config.allow_nested,
                    allow_io_errors=self.config.allow_io_errors
                )
            for core_path in locator.locate_core_paths():
                site = WordpressSite(core_path)
                log.debug(
                        'Located WordPress site at ' + os.fsdecode(core_path)
                    )
                try:
                    database = site.get_database()
                    yield database
                except WordpressException:
                    if self.config.allow_io_errors:
                        log.warning(
                                'Failed to extract database credentials '
                                'for site at ' + os.fsdecode(core_path)
                            )
                    else:
                        raise

    def _get_json_validator(self) -> ListValidator:
        return ListValidator(
                DictionaryValidator({
                        'name': str,
                        'user': str,
                        'password': str,
                        'host': str,
                        'port': OptionalValueValidator(int),
                        'collation': OptionalValueValidator(str),
                        'prefix': OptionalValueValidator(str)
                    }, optional_keys={'port', 'collation'})
            )

    def _parse_configured_databases(
                self,
                io_manager: IoManager
            ) -> Generator[WordpressDatabase, None, None]:
        validator = self._get_json_validator()
        for path in self._get_search_paths(io_manager):
            with open(path, 'rb') as file:
                configList = json.load(file)
                validator.validate(configList)
                for config in configList:
                    try:
                        port = config['port']
                    except KeyError:
                        port = DEFAULT_PORT
                    try:
                        collation = config['collation']
                    except KeyError:
                        collation = DEFAULT_COLLATION
                    try:
                        prefix = config['prefix']
                    except KeyError:
                        prefix = DEFAULT_PREFIX
                    yield WordpressDatabase(
                            name=config['name'],
                            server=WordpressDatabaseServer(
                                    host=config['host'],
                                    port=port,
                                    user=config['user'],
                                    password=config['password']
                                ),
                            prefix=prefix,
                            collation=collation
                        )

    def _get_databases(
                self,
                io_manager: IoManager
            ) -> List[WordpressDatabase]:
        databases = []
        base = self._get_base_database()
        if base is not None:
            databases.append(base)
        generator = self._locate_site_databases(io_manager) if \
            self.config.locate_sites else \
            self._parse_configured_databases(io_manager)
        for database in generator:
            databases.append(database)
        return databases

    def _load_remote_rules(self) -> DatabaseRuleSet:

        def fetch_rules() -> DatabaseRuleSet:
            client = self.context.get_noc1_client()
            return client.get_database_rules()

        cacheable = caching.Cacheable(
                'database_rules',
                fetch_rules,
                caching.DURATION_ONE_DAY
            )

        return cacheable.get(self.cache)

    def _filter_rules(self, rule_set: DatabaseRuleSet) -> None:
        included = None
        if self.config.include_rules:
            included = set(self.config.include_rules)
        excluded = None
        if self.config.exclude_rules:
            excluded = set(self.config.exclude_rules)
        rule_set.filter_rules(included, excluded)

    def _load_rules(self) -> DatabaseRuleSet:
        rule_set = self._load_remote_rules() \
            if self.config.use_remote_rules \
            else DatabaseRuleSet()
        if self.config.rules_file is not not_set_token:
            for rules_file in self.config.rules_file:
                load_database_rules(rules_file, rule_set)
        self._filter_rules(rule_set)
        return rule_set

    def invoke(self) -> int:
        report_manager = DatabaseScanReportManager(self.context)
        io_manager = report_manager.get_io_manager()
        rule_set = self._load_rules()
        scanner = DatabaseScanner(rule_set)
        with report_manager.open_output_file() as output_file:
            report = report_manager.initialize_report(output_file)
            for database in self._get_databases(io_manager):
                for result in scanner.scan(database):
                    report.add_result(result)
            report.database_count = scanner.scan_count
            report.complete()
        if self.context.requires_input(self.config.require_database) \
                and scanner.scan_count == 0:
            raise ConfigurationException(
                    'At least one database to scan must be specified'
                )
        elapsed_time = round(scanner.get_elapsed_time())
        log.info(
                f'Found {report.result_count} result(s) after scanning '
                f'{scanner.scan_count} database(s) over {elapsed_time} '
                'second(s)'
            )
        return 0


factory = DbScanSubcommand



import signal
from typing import Optional

from ...intel.signatures import CommonString, Signature, SignatureSet
from ...logging import log
from ...util import pcre
from ...util.pcre import PcrePattern, PcreException, PcreJitStack, \
        PcreOptions, PCRE_DEFAULT_OPTIONS

from .matching import Matcher, BaseMatcherContext, TimeoutException, \
        MatchWorkspace, MatchEngineCompilerOptions, MatchEngineOptions, \
        DEFAULT_TIMEOUT


if not pcre.AVAILABLE:
    raise RuntimeError('PCRE is not available')


class PcreCommonString:

    def __init__(self, common_string: CommonString, pcre_options: PcreOptions):
        self.common_string = common_string
        self.pattern = PcrePattern(common_string.string, pcre_options)


class PcreSignature:

    def __init__(self, signature: Signature, pcre_options: PcreOptions):
        self.signature = signature
        self.anchored_to_start = self._is_anchored_to_start()
        self.pcre_options = pcre_options
        if not signature.has_common_strings():
            self.compile()

    def _is_anchored_to_start(self) -> bool:
        try:
            first_character = self.signature.rule[0]
            return first_character == '^'
        except IndexError:
            # Patterns shouldn't be empty, but if they are, they're not
            # anchored
            return False

    def is_valid(self) -> bool:
        return self.get_pattern() is not None

    def compile(self) -> None:
        try:
            rule = self.signature.rule
            self.pattern = PcrePattern(rule, self.pcre_options)
        except BaseException as error:  # noqa: B036
            log.error('Regex compilation for signature ' +
                      str(self.signature.identifier) +
                      ' failed: ' +
                      str(error) +
                      ', pattern: ' +
                      repr(rule))
            self.pattern = None

    def get_pattern(self) -> PcrePattern:
        # Signature patterns are compiled lazily as they are only needed if
        # common strings are matched and compiling all takes several seconds
        if not hasattr(self, 'pattern'):
            self.compile()
        return self.pattern


class PcreMatchWorkspace(MatchWorkspace):

    def __init__(self):
        self.jit_stack = None

    def __enter__(self):
        self.jit_stack = PcreJitStack()
        return self

    def __exit__(self, exc_type, exc_value, traceback) -> None:
        self.jit_stack.__exit__(exc_type, exc_value, traceback)


class PcreMatcherContext(BaseMatcherContext):

    def __init__(self, matcher: Matcher):
        super().__init__(matcher)
        self.common_string_states = self._initialize_common_string_states()

    def _initialize_common_string_states(self) -> list:
        states = []
        for _common_string in self.matcher.common_strings:
            states.append(False)
        return states

    def _check_common_strings(self, chunk: bytes) -> list:
        common_string_counts = {}
        for index, common_string in enumerate(self.matcher.common_strings):
            if not self.common_string_states[index]:
                try:
                    match = common_string.pattern.match(chunk)
                    if match is not None:
                        self.common_string_states[index] = True
                except PcreException as e:
                    log.debug(
                            'Common string matching failed for '
                            f'{common_string.common_string.string} with error:'
                            f' {e}'
                        )
            if self.common_string_states[index]:
                for identifier in common_string.common_string.signature_ids:
                    if identifier in self.matches:
                        continue
                    if identifier in common_string_counts:
                        common_string_counts[identifier] += 1
                    else:
                        common_string_counts[identifier] = 1
        possible_signatures = []
        for identifier, count in common_string_counts.items():
            signature = self.matcher.signatures[identifier]
            if count == signature.signature.get_common_string_count():
                possible_signatures.append(signature)
        return possible_signatures

    def _match_signature(
                self,
                signature: Signature,
                chunk: bytes,
                start: bool = False,
                jit_stack: Optional[PcreJitStack] = None
            ) -> bool:
        if not signature.is_valid():
            print('Signature is not valid')
            return False
        if signature.anchored_to_start and not start:
            return False
        try:
            signal.setitimer(signal.ITIMER_VIRTUAL, self.matcher.timeout)
            match = signature.get_pattern().match(chunk, jit_stack)
            signal.setitimer(signal.ITIMER_VIRTUAL, 0)  # Clear timeout
            if match is not None:
                self._record_match(
                        identifier=signature.signature.identifier,
                        matched=match.matched_string
                    )
                return True
        except PcreException as e:
            log.debug(
                    'Signature matching failed for '
                    f'{signature.signature.identifier}, {e}'
                )
        except TimeoutException:
            self.timeouts.add(signature.signature.identifier)
        return False

    def process_chunk(
                self,
                chunk: bytes,
                start: bool = False,
                workspace: Optional[MatchWorkspace] = None
            ) -> bool:
        possible_signatures = self._check_common_strings(chunk)
        for signature in self.matcher.signatures_without_common_strings:
            if self._match_signature(
                        signature, chunk, start, workspace.jit_stack
                    ) and not self.matcher.match_all:
                return True
        for signature in possible_signatures:
            if self._match_signature(
                        signature, chunk, start, workspace.jit_stack
                    ) and not self.matcher.match_all:
                return True
        return False

    def __enter__(self):
        def handle_timeout(signum, frame):
            raise TimeoutException()

        self._previous_alarm_handler = signal.signal(
                signal.SIGVTALRM,
                handle_timeout
            )
        if self._previous_alarm_handler is None:
            self._previous_alarm_handler = signal.SIG_DFL
        return super().__enter__()

    def __exit__(self, exc_type, exc_value, traceback) -> None:
        signal.signal(signal.SIGVTALRM, self._previous_alarm_handler)


class PcreMatcher(Matcher):

    def __init__(
                self,
                signature_set: SignatureSet,
                timeout: int = DEFAULT_TIMEOUT,
                match_all: bool = False,
                pcre_options: PcreOptions = PCRE_DEFAULT_OPTIONS,
                lazy: bool = False
            ):
        self.pcre_options = pcre_options
        super().__init__(signature_set, timeout, match_all, lazy)

    def _extract_signatures_without_common_strings(self) -> list:
        signatures = []
        for signature in self.signatures.values():
            if not signature.signature.has_common_strings():
                signatures.append(signature)
        return signatures

    def _compile_common_strings(self) -> None:
        self.common_strings = [
                PcreCommonString(common_string, self.pcre_options)
                for common_string in self.signature_set.common_strings
            ]

    def _compile_signatures(self) -> None:
        self.signatures = {}
        for identifier, signature in self.signature_set.signatures.items():
            self.signatures[identifier] = PcreSignature(
                    signature,
                    self.pcre_options
                )

    def _compile_regexes(self) -> None:
        self._compile_common_strings()
        self._compile_signatures()
        self.signatures_without_common_strings = \
            self._extract_signatures_without_common_strings()

    def _prepare(self) -> None:
        self._compile_regexes()

    def create_context(self) -> PcreMatcherContext:
        return PcreMatcherContext(self)

    def create_workspace(self) -> PcreMatchWorkspace:
        return PcreMatchWorkspace()


def create_compiler(options: MatchEngineCompilerOptions) -> None:
    return None


def validate_pre_compiled_data(source: bytes) -> bool:
    return False  # Pre-compilation is not supported


def create_matcher(options: MatchEngineOptions) -> PcreMatcher:
    return PcreMatcher(
            options.signature_set,
            match_all=options.match_all,
            pcre_options=options.pcre_options,
            lazy=options.lazy
        )



from dataclasses import dataclass
from enum import Enum
from typing import List, Optional
from importlib import import_module
from contextlib import AbstractContextManager

from ...intel.signatures import SignatureSet
from ...util.caching import Cacheable
from ...util.pcre import PcreOptions, PCRE_DEFAULT_OPTIONS

DEFAULT_TIMEOUT = 1  # Seconds


class TimeoutException(Exception):
    pass


class MatchResult:

    def __init__(self, matches: list):
        self.matches = matches

    def matches(self) -> bool:
        return len(self.matches) > 0


class MatchWorkspace(AbstractContextManager):

    def __enter__(self):
        return self

    def __exit__(self, exc_type, exc_value, traceback) -> None:
        return self


class MatcherContext(AbstractContextManager):

    def __init__(self):
        self.matches = {}
        self.timeouts = set()

    def process_chunk(
                self,
                chunk: bytes,
                start: bool = False,
                workspace: Optional[MatchWorkspace] = None
            ):
        raise NotImplementedError()

    def finalize_content(self) -> None:
        pass

    def _record_match(self, identifier: str, matched: Optional[bytes]) -> None:
        self.matches[identifier] = matched

    def __enter__(self):
        return self

    def __exit__(self, exc_type, exc_value, traceback) -> None:
        pass


class Matcher:

    def __init__(
                self,
                signature_set: SignatureSet,
                timeout: int = DEFAULT_TIMEOUT,
                match_all: bool = False,
                lazy: bool = False
            ):
        self.signature_set = signature_set
        self.timeout = timeout
        self.match_all = match_all
        self.prepared = False
        self.prepared_thread = False
        if not lazy:
            self.prepare()

    def get_cacheable(self) -> Optional[Cacheable]:
        return None

    def prepare(self, thread: bool = False) -> None:
        if not self.prepared:
            self._prepare()
            self.prepared = True
        if thread:
            self.prepare_thread()

    def _prepare(self) -> None:
        pass

    def prepare_thread(self) -> None:
        if self.prepared_thread:
            return
        self._prepare_thread()
        self.prepared_thread = True

    def _prepare_thread(self) -> None:
        pass

    def create_workspace(self) -> Optional[MatchWorkspace]:
        return MatchWorkspace()

    def create_context(self) -> MatcherContext:
        raise NotImplementedError()


class BaseMatcherContext(MatcherContext):

    def __init__(self, matcher: Matcher):
        self.matcher = matcher
        super().__init__()


@dataclass
class MatchEngineCompilerOptions:
    generic: bool = False


@dataclass
class MatchEngineOptions:
    signature_set: SignatureSet
    match_all: bool = False
    lazy: bool = False
    pcre_options: PcreOptions = PCRE_DEFAULT_OPTIONS
    database_source: Optional[bytes] = None


class Compiler:

    def compile_serializable(self, signatures: SignatureSet) -> bytes:
        raise NotImplementedError()


class MatchEngine(Enum):
    PCRE = 'pcre'
    VECTORSCAN = 'vectorscan'

    def __init__(self, option: str, module: Optional[str] = None):
        self.option = option
        self.module = option if module is None else module
        self._loaded_module = None

    @classmethod
    def get_options(cls) -> List:
        return [engine.option for engine in cls]

    @classmethod
    def for_option(cls, option: str):
        for engine in cls:
            if engine.option == option:
                return engine
        raise ValueError(f'Unrecognized engine option: {option}')

    @classmethod
    def get_default(cls):
        return cls.PCRE

    @classmethod
    def get_default_option(cls):
        return cls.get_default().option

    def _load_module(self):
        return import_module(
                f'.{self.module}',
                'wordfence.scanning.matching'
            )

    def _get_loaded_module(self):
        if self._loaded_module is None:
            self._loaded_module = self._load_module()
        return self._loaded_module

    def get_compiler(
                self,
                options: MatchEngineCompilerOptions
            ) -> Optional[Compiler]:
        module = self._get_loaded_module()
        return module.create_compiler(options)

    def supports_pre_compilation(self) -> bool:
        compiler = self.get_compiler(MatchEngineCompilerOptions())
        return compiler is not None

    def validate_database_source(self, source: Optional[bytes]) -> bool:
        if source is None:
            return True
        module = self._get_loaded_module()
        return module.validate_database_source(source)

    def create_matcher(self, options: MatchEngineOptions) -> Matcher:
        module = self._get_loaded_module()
        return module.create_matcher(options)



import sys
import logging
from typing import Set

from ..util import updater
from ..util.terminal import supports_colors
from ..logging import log
from ..scanning.scanner import ExceptionContainer
from .banner.banner import show_welcome_banner_if_enabled
from .config import load_config, RenamedSubcommandException, GlobalConfig
from .config.base_config_definitions import config_map \
        as base_config_map
from .subcommands import load_subcommand_definitions
from .context import CliContext
from .configurer import Configurer
from . import licensing
from .terms_management import TermsManager
from .import terms_management
from .helper import Helper


class ExceptionHandler:

    def __init__(self):
        self.global_config = GlobalConfig()
        self.subcommand = None

    def print_error(self, message: str) -> None:
        if sys.stderr is not None:
            print(message, file=sys.stderr)
        else:
            print(message)

    def process_exception(self, exception: BaseException) -> int:
        if self.subcommand is not None:
            self.subcommand.terminate()
        if isinstance(exception, ExceptionContainer):
            if self.global_config.debug:
                self.print_error(exception.trace)
                return 1
            exception = exception.exception
        if self.global_config.debug:
            raise exception
        else:
            if isinstance(exception, SystemExit):
                raise exception
            else:
                if self.subcommand is None:
                    message = None
                else:
                    message = self.subcommand.generate_exception_message(
                            exception
                        )
                if message is None:
                    message = f'Error: {exception}'
                self.print_error(message)
        return 1


class WordfenceCli:

    def __init__(self, exception_handler: ExceptionHandler):
        self.exception_handler = exception_handler
        self.initialize_early_logging()
        self.subcommand_definitions = load_subcommand_definitions()
        self.helper = self._initialize_helper()
        self._load_config(exception_handler)
        self.allows_color = self.config.color is not False \
            and supports_colors()

    def _initialize_helper(self) -> Helper:
        return Helper(
                self.subcommand_definitions,
                base_config_map
            )

    def _load_config(self, exception_handler: ExceptionHandler) -> None:
        try:
            self.config, self.subcommand_definition = load_config(
                    self.subcommand_definitions,
                    self.helper,
                    global_config=exception_handler.global_config
                )
        except RenamedSubcommandException as rename:
            print(
                    f'The "{rename.old}" subcommand has been renamed to '
                    f'"{rename.new}"'
                )
            sys.exit(1)

    def initialize_early_logging(self) -> None:
        log.setLevel(logging.INFO)

    def _get_cacheable_types(self) -> Set[str]:
        cacheable_types = set()
        cacheable_types.update(licensing.CACHEABLE_TYPES)
        cacheable_types.update(terms_management.CACHEABLE_TYPES)
        for definition in self.subcommand_definitions.values():
            cacheable_types.update(definition.cacheable_types)
        return cacheable_types

    def display_help(self) -> None:
        self.helper.display_help(self.config.subcommand)

    def configure_stdio(self) -> bool:
        original_encoding = sys.stdout.encoding
        sys.stdout.reconfigure(encoding='utf-8', errors='ignore')
        if original_encoding != 'utf-8':
            log.warning(
                    f'Encoding for stdout is {original_encoding} instead of '
                    'utf-8'
                )
            return True
        else:
            return False

    def invoke(self) -> int:
        with CliContext(
                    self.config,
                    self._get_cacheable_types(),
                    self.helper,
                    self.allows_color
                ) as context:
            context.initialize_logging()
            stdio_changed = self.configure_stdio()

            if self.config.purge_cache:
                context.cache.purge()

            if not stdio_changed:
                show_welcome_banner_if_enabled(self.config)

            if self.config.help:
                self.display_help()
                return 0

            if self.config.version:
                context.display_version()
                return 0

            if self.config.check_for_update:
                updater.Version.check(context.cache)

            license_manager = licensing.LicenseManager(context)
            context.register_license_update_hook(
                    license_manager.update_license
                )

            terms_manager = TermsManager(context, license_manager)
            context.register_terms_update_hook(terms_manager.trigger_update)

            configurer = Configurer(
                    context,
                    self.helper,
                    license_manager,
                    terms_manager,
                    self.subcommand_definitions,
                    self.subcommand_definition
                )
            context.configurer = configurer

            if self.subcommand_definition is None:
                self.display_help()
                configurer.check_config()
                return 0

            if self.subcommand_definition.requires_config:
                if not configurer.check_config():
                    return 0
                if not self.subcommand_definition.uses_license:
                    license_manager.check_license()
                terms_manager.prompt_acceptance_if_needed()

            subcommand = self.subcommand_definition.initialize_subcommand(
                    context
                )
            self.exception_handler.subcommand = subcommand
            return subcommand.invoke()


def invoke_cli():
    exception_handler = ExceptionHandler()
    try:
        cli = WordfenceCli(exception_handler)
        return cli.invoke()
    except BaseException as exception:  # noqa: B036
        return exception_handler.process_exception(exception)
    except KeyboardInterrupt:
        return 130


def main():
    exit_code = invoke_cli()
    sys.exit(exit_code)


if __name__ == '__main__':
    main()



from ..subcommands import SubcommandDefinition, UsageExample
from ..config.typing import ConfigDefinitions

config_definitions: ConfigDefinitions = {
    "read-stdin": {
        "description": "Read paths from stdin. If not specified, paths will "
                       "automatically be read from stdin when input is not "
                       "from a TTY.",
        "context": "ALL",
        "argument_type": "OPTIONAL_FLAG",
        "default": None
    },
    "path-separator": {
        "short_name": "s",
        "description": "Separator used to delimit paths when reading from "
                       "stdin. Defaults to the null byte.",
        "context": "ALL",
        "argument_type": "OPTION",
        "default": "AA==",
        "default_type": "base64"
    },
    "require-path": {
        "description": "When enabled, invoking the count command without "
                       "specifying at least one path will trigger an error. "
                       "This is the default behavior when running in a "
                       "terminal.",
        "context": "CLI",
        "argument_type": "OPTIONAL_FLAG",
        "default": None
    },
    "allow-nested": {
        "description": "When enabled (the default), WordPress installations "
                       "nested below other installations will be included in "
                       "the count.",
        "context": "ALL",
        "argument_type": "FLAG",
        "default": True
    },
    "allow-io-errors": {
        "description": "Allow counting to continue if IO errors are "
                       "encountered. Sites that cannot be identified due to "
                       "IO errors will be omitted from the count. This is the "
                       "default behavior.",
        "context": "ALL",
        "argument_type": "FLAG",
        "default": True
    }
}

examples = [
    UsageExample(
        'Count the number of WordPress installations under /var/www/',
        'wordfence count-sites /var/www/'
    )
]

definition = SubcommandDefinition(
    name='count-sites',
    usage='[OPTIONS] [PATH]...',
    description='Count the total number of WordPress installations under '
                'the specified path(s)',
    config_definitions=config_definitions,
    config_section='COUNT_SITES',
    cacheable_types=set(),
    examples=examples,
    accepts_directories=True
)



from typing import List, Optional, Dict

from wordfence.databasescanning.scanner import DatabaseScanResult
from wordfence.util.terminal import Color, escape, RESET
from wordfence.util.json import safe_json_encode
from ..reporting import ReportManager, ReportColumnEnum, ReportFormatEnum, \
    ReportRecord, Report, ReportFormat, ReportColumn, ReportEmail, \
    BaseHumanReadableWriter, \
    get_config_options, generate_html_table, generate_report_email_html, \
    REPORT_FORMAT_CSV, REPORT_FORMAT_TSV, REPORT_FORMAT_NULL_DELIMITED, \
    REPORT_FORMAT_LINE_DELIMITED
from ..context import CliContext
from ..email import Mailer


class DatabaseScanReportColumn(ReportColumnEnum):
    TABLE = 'table', lambda record: record.result.table
    RULE_ID = 'rule_id', lambda record: record.result.rule.identifier
    RULE_DESCRIPTION = 'rule_description', \
        lambda record: record.result.rule.description
    ROW = 'row', lambda record: safe_json_encode(record.result.row)


class HumanReadableWriter(BaseHumanReadableWriter):

    def format_record(self, record) -> str:
        result = record.result
        return (
                escape(Color.YELLOW)
                + 'Suspicious database record found in table '
                f'"{result.table}" matching rule "{result.rule.description}"'
                ': ' + safe_json_encode(record.result.row) + RESET
            )


REPORT_FORMAT_HUMAN = ReportFormat(
        'human',
        lambda stream, columns: HumanReadableWriter(stream),
        allows_headers=False,
        allows_column_customization=False
    )


class DatabaseScanReportFormat(ReportFormatEnum):
    CSV = REPORT_FORMAT_CSV
    TSV = REPORT_FORMAT_TSV
    NULL_DELIMITED = REPORT_FORMAT_NULL_DELIMITED
    LINE_DELIMITED = REPORT_FORMAT_LINE_DELIMITED
    HUMAN = REPORT_FORMAT_HUMAN


class DatabaseScanReportRecord(ReportRecord):

    def __init__(self, result: DatabaseScanResult):
        self.result = result


class DatabaseScanReport(Report):

    def __init__(
                self,
                format: ReportFormat,
                columns: List[ReportColumn],
                email_addresses: List[str],
                mailer: Optional[Mailer],
                write_headers: bool = False,
                only_unremediated: bool = False
            ):
        super().__init__(
                format,
                columns,
                email_addresses,
                mailer,
                write_headers
            )
        self.result_count = 0
        self.database_count = 0

    def add_result(self, result: DatabaseScanResult):
        self.result_count += 1
        self.write_record(
                DatabaseScanReportRecord(result)
            )

    def generate_email(
                self,
                recipient: str,
                attachments: Dict[str, str],
                hostname: str
            ) -> ReportEmail:
        plain = (
                'Database Scan Complete\n\n'
                f'Scanned Databases: {self.database_count}\n\n'
                f'Results Found: {self.result_count}\n\n'
            )

        results = {
                'Scanned Databases': self.database_count,
                'Results Found': self.result_count
            }

        table = generate_html_table(results)

        document = generate_report_email_html(
                table,
                'Database Scan Results',
                hostname
            )

        return ReportEmail(
                recipient=recipient,
                subject=f'Database Scan Results for {hostname}',
                plain_content=plain,
                html_content=document.to_html()
            )


class DatabaseScanReportManager(ReportManager):

    def __init__(self, context: CliContext):
        super().__init__(
                formats=DatabaseScanReportFormat,
                columns=DatabaseScanReportColumn,
                context=context,
                read_stdin=context.config.read_stdin,
                input_delimiter=context.config.path_separator,
                binary_input=True
            )

    def _instantiate_report(
                self,
                format: ReportFormat,
                columns: List[ReportColumn],
                email_addresses: List[str],
                mailer: Optional[Mailer],
                write_headers: bool
            ) -> Report:
        return DatabaseScanReport(
                format,
                columns,
                email_addresses,
                mailer,
                write_headers
            )


DATABASE_SCAN_REPORT_CONFIG_OPTIONS = get_config_options(
        DatabaseScanReportFormat,
        DatabaseScanReportColumn,
        default_format='human'
    )



from .noc_client import NocClient

NOC4_BASE_URL = 'https://noc4.wordfence.com/v1.11/'


class Client(NocClient):

    def get_default_base_url(self) -> str:
        return NOC4_BASE_URL

    def build_query(self, action: str, base_query: dict = None) -> dict:
        query = super().build_query(action, base_query)
        # TODO: How should site parameters be handled for CLI requests
        query['s'] = 'http://www.example.com'
        query['h'] = 'http://www.example.com'
        return query



import signal


HANDLED_SIGNALS = [
        signal.SIGINT
    ]


_handlers = None


def reset():
    global _handlers
    if _handlers is None:
        _handlers = {}
        for signal_type in HANDLED_SIGNALS:
            _handlers[signal_type] = signal.getsignal(signal_type)
            signal.signal(signal_type, signal.SIG_DFL)


def restore():
    global _handlers
    for signal_type, handler in _handlers.items():
        signal.signal(signal_type, handler)
    _handlers = None



from ..subcommands import SubcommandDefinition
from ..config.typing import ConfigDefinitions

config_definitions: ConfigDefinitions = {}

cacheable_types = set()

definition = SubcommandDefinition(
    name='help',
    usage='[SUBCOMMAND]',
    description='Display help',
    config_definitions=config_definitions,
    config_section='DEFAULT',
    cacheable_types=cacheable_types,
    requires_config=False
)



import re
import os
from typing import Optional, Dict, List

from ..logging import log
from ..util.io import SYMLINK_IO_ERRORS
from ..util.encoding import str_to_bytes
from .exceptions import ExtensionException


HEADER_READ_SIZE = 8 * 1024
HEADER_CLEANUP_PATTERN = re.compile(r'\s*(?:\*\/|\?>).*')
CARRIAGE_RETURN_PATTERN = re.compile('\r')


class Extension:

    def __init__(
                self,
                slug: str,
                version: Optional[bytes],
                header: Dict[str, str],
                path: bytes
            ):
        self.slug = slug
        self.version = version
        self.header = header
        self.path = path

    def get_name(self) -> str:
        try:
            return self.header['Name']
        except KeyError:
            return self.slug

    def __str__(self) -> str:
        return f'{self.slug}({self.version})'


class ExtensionLoader:

    def __init__(
                self,
                extension_type: str,
                directory: str,
                header_fields: Dict[str, str],
                allow_io_errors: bool = False,
            ):
        self.extension_type = extension_type
        self.directory = directory
        self.header_fields = header_fields
        self.allow_io_errors = allow_io_errors

    def _clean_up_header_value(self, value: str) -> str:
        return HEADER_CLEANUP_PATTERN.sub('', value).strip()

    def _read_header(self, path: bytes) -> str:
        try:
            with open(path, 'r', errors='replace') as stream:
                data = stream.read(HEADER_READ_SIZE)
                return data
        except OSError as error:
            raise ExtensionException(
                    f'Unable to read {self.extension_type} header from '
                    + os.fsdecode(path)
                ) from error

    def _parse_header(
                self,
                data: str,
            ) -> Dict[str, str]:
        data = CARRIAGE_RETURN_PATTERN.sub('\n', data)
        values = {}
        for field, pattern in self.header_fields.items():
            match = re.search(
                    '^[ \t\\/*#@]*' + re.escape(pattern) + r':(.*)$',
                    data,
                    re.MULTILINE | re.IGNORECASE
                )
            if match is not None:
                values[field] = self._clean_up_header_value(match.group(1))
        return values

    def load(
                self,
                slug: str,
                path: bytes,
                base_path: Optional[bytes] = None
            ) -> Optional[Extension]:
        header_data = self._read_header(path)
        header = self._parse_header(header_data)
        if 'Name' not in header:
            return None
        try:
            version = header['Version']
            if isinstance(version, str):
                version = str_to_bytes(version)
        except KeyError:
            version = None
        if base_path is None:
            base_path = path
        return self._initialize_extension(slug, version, header, base_path)

    def _initialize_extension(
                self,
                slug: str,
                version: Optional[str],
                header: Dict[str, str],
                path: bytes
            ):
        return Extension(
                slug=slug,
                version=version,
                header=header,
                path=path
            )

    def _process_entry(entry: os.DirEntry) -> Optional[Extension]:
        return None

    def load_all(self) -> List[Extension]:
        extensions = []
        try:
            for entry in os.scandir(self.directory):
                try:
                    extension = self._process_entry(entry)
                    if extension is not None:
                        extensions.append(extension)
                except OSError as error:
                    if error.errno in SYMLINK_IO_ERRORS:
                        continue
                    if self.allow_io_errors:
                        log.warning(
                                f'Unable to load {self.extension_type} from '
                                + os.fsdecode(entry.path) + f': {error}'
                            )
                    else:
                        raise
        except OSError as error:
            if error.errno not in SYMLINK_IO_ERRORS:
                message = (
                        f'Unable to scan {self.extension_type} directory at '
                        + os.fsdecode(self.directory)
                    )
                if self.allow_io_errors:
                    log.warning(message)
                else:
                    raise ExtensionException(message) from error
        return extensions



import re
from typing import List, Dict, Union, Optional

from .encoding import str_to_bytes, bytes_to_str


PHP_VERSION_DELIMITER = b'.'
PHP_VERSION_ALTERNATE_DELIMITERS = [b'_', b'-', b'+']


NON_NUMBER_PATTERN = re.compile(b'[^0-9.]+')
NUMBER_PATTERN = re.compile(b'^[0-9]+$')
REPEATED_DOT_PATTERN = re.compile(b'\\.{2,}')


# TODO: Convert str to bytes
def delimit_non_numbers(version: str) -> str:
    return NON_NUMBER_PATTERN.sub(b".\\g<0>.", version).strip(b'.')


def is_number(string: bytes) -> bool:
    return NUMBER_PATTERN.match(string) is not None


def strip_repeated_delimiters(version: bytes) -> bytes:
    return REPEATED_DOT_PATTERN.sub(b'.', version)


LOWER_ALPHA_VERSIONS = [
    [b'dev'],
    [b'alpha', b'a'],
    [b'beta', b'b'],
    [b'RC', b'rc'],
]
HIGHER_ALPHA_VERSIONS = [
    [b'pl', b'p']
]
TIER_OFFSET = 2
TIER_NUMBER = len(LOWER_ALPHA_VERSIONS) + TIER_OFFSET


def create_alpha_version_map(versions: List[List[bytes]]) -> Dict[bytes, int]:
    map = {}
    for index, tier in enumerate(versions):
        for version in tier:
            map[version] = index
    return map


LOWER_ALPHA_VERSION_MAP = create_alpha_version_map(LOWER_ALPHA_VERSIONS)
HIGHER_ALPHA_VERSIONS_MAP = create_alpha_version_map(HIGHER_ALPHA_VERSIONS)


def get_alpha_tier(string: bytes, map: Dict[bytes, int]) -> Optional[int]:
    try:
        return map[string]
    except KeyError:
        return None


def get_lower_alpha_tier(string: bytes) -> Optional[int]:
    return get_alpha_tier(string, LOWER_ALPHA_VERSION_MAP)


def get_higher_alpha_tier(string: bytes) -> Optional[int]:
    return get_alpha_tier(string, HIGHER_ALPHA_VERSIONS_MAP)


class PhpVersionComponent:

    def __init__(self, value: bytes):
        self.is_number = is_number(value)
        if self.is_number:
            self.value = int(value)
        else:
            self.value = value
        self.lower_alpha_tier = \
            None if self.is_number else get_lower_alpha_tier(self.value)
        self.higher_alpha_tier = \
            None if self.is_number else get_higher_alpha_tier(self.value)
        self.tier = self._evaluate_tier()

    def _evaluate_tier(self) -> int:
        if self.lower_alpha_tier is not None:
            return self.lower_alpha_tier + TIER_OFFSET
        if self.is_number:
            return TIER_NUMBER
        if self.higher_alpha_tier is not None:
            return TIER_NUMBER + 1 + self.higher_alpha_tier
        if not self.is_number and self.lower_alpha_tier is None \
                and self.higher_alpha_tier is None:
            return 1
        return 0

    def __str__(self) -> str:
        return str(self.value)


DefaultComponent = PhpVersionComponent(b'0')


class PhpVersion:

    def __init__(self, version: Union[str, bytes]):
        if isinstance(version, str):
            version = str_to_bytes(version)
        self.version = version
        self._components = self.extract_components(version)

    def extract_components(self, version: bytes) -> List[bytes]:
        for character in PHP_VERSION_ALTERNATE_DELIMITERS:
            version = version.replace(character, PHP_VERSION_DELIMITER)
        # Note that this also strips leading/trailing delimiters
        version = strip_repeated_delimiters(
                delimit_non_numbers(version)
            )
        return list(map(PhpVersionComponent, version.split(b'.')))

    def _get_component(self, index: int) -> PhpVersionComponent:
        try:
            return self._components[index]
        except IndexError:
            return DefaultComponent


def compare_version_components(
            a: Optional[PhpVersionComponent],
            b: Optional[PhpVersionComponent]
        ) -> int:
    if a.value == b.value:
        return 0
    if a.tier != b.tier:
        return -1 if a.tier < b.tier else 1
    if a.tier == 0 or a.tier == TIER_NUMBER:
        return -1 if a.value < b.value else 1
    return 0


def compare_php_versions(
            a: Union[PhpVersion, str, bytes],
            b: Union[PhpVersion, str, bytes]
        ) -> int:
    """ This is intended to mirror PHP's version_compare function  """
    """ https://www.php.net/manual/en/function.version-compare.php """
    if not isinstance(a, PhpVersion):
        a = PhpVersion(a)
    if not isinstance(b, PhpVersion):
        b = PhpVersion(b)
    component_count = max(len(a._components), len(b._components))
    for i in range(0, component_count):
        comparison = compare_version_components(
                a._get_component(i),
                b._get_component(i)
            )
        if comparison != 0:
            return comparison
    return 0


def version_to_str(version: Optional[bytes]) -> str:
    if version is None:
        return 'unknown'
    return bytes_to_str(version)



import os
import sys

TEXT_BANNER = r"""
 _       __               __  ____
| |     / /___  _________/ / / __/__  ____  ________
| | /| / / __ \/ ___/ __  /_/ /_/ _ \/ __ \/ ___/ _ \
| |/ |/ / /_/ / /  / /_/ /_  __/  __/ / / / /__/  __/
|__/|__/\____/_/   \____/ /_/  \___/_/ /_/\___/\___/
                                      ____ _     ___
                                     / ___| |   |_ _|
                                    | |   | |    | |
                                    | |___| |___ | |
                                     \____|_____|___|
"""

LOGO = r"""
             ▓▓▓
        ▓▓▓▓▓   ▓▓▓▓▓
  ▓▓▓▓▓▓▓           ▓▓▓▓▓▓▓
 ▓▓           ▓           ▓▓
▓▓     ▓▓    ▓▓▓    ▓▓     ▓▓
▓▓      ▓     ▓     ▓      ▓▓
▓▓    ▓▓▓▓▓▓▓▓▓▓▓▓▓▓▓▓▓    ▓▓
▓▓  ▓▓▓ ▓    ▓▓▓    ▓ ▓▓▓  ▓▓
▓▓▓▓▓   ▓    ▓▓▓    ▓   ▓▓▓▓▓
 ▓▓     ▓   ▓▓ ▓▓   ▓     ▓▓
  ▓▓▓▓▓▓▓▓▓▓▓▓ ▓▓▓▓▓▓▓▓▓▓▓▓
"""


class Banner:

    def __init__(self, content: str):
        self.content = content
        self.process_content()

    def process_content(self) -> None:
        self.row_count = 0
        self.column_count = 0
        rows = self.content.split('\n')
        for row in rows:
            self.column_count = max(self.column_count, len(row.rstrip()))
            self.row_count += 1
        for index, row in enumerate(rows):
            rows[index] = row.ljust(self.column_count)
        self.rows = rows

    def merge(self, banner, separator: str = ' ') -> None:
        height_difference = self.row_count - banner.row_count
        self_taller = height_difference > 0
        taller = self if self_taller else banner
        if self_taller:
            self_offset = 0
            banner_offset = -height_difference
        else:
            self_offset = -height_difference
            banner_offset = 0
        height_difference = abs(height_difference)
        new_rows = []
        for index in range(0, height_difference):
            new_rows.append(taller.rows[index])
        for index in range(height_difference, taller.row_count):
            new_rows.append(
                    self.rows[index + self_offset] +
                    separator +
                    banner.rows[index + banner_offset]
                )
        self.rows = new_rows
        self.row_count += height_difference
        self.column_count += len(separator) + banner.column_count

    def display(self) -> None:
        for row in self.rows:
            print(row)

    def __str__(self) -> str:
        return self.content


def add_logo(banner) -> str:
    pass


def get_welcome_banner():
    terminal_columns = os.get_terminal_size().columns
    text = Banner(TEXT_BANNER)
    logo = Banner(LOGO)
    combined = Banner(LOGO)
    combined.merge(text)
    variants = [
            combined,
            text,
            logo
        ]
    for banner in variants:
        if banner.column_count <= terminal_columns:
            return banner
    return None


def show_welcome_banner():
    banner = get_welcome_banner()
    if banner is not None:
        banner.display()


def should_show_welcome_banner(banner_enabled):
    return banner_enabled \
            and sys.stdout.isatty() \
            and sys.stdout.encoding == 'utf-8'


def show_welcome_banner_if_enabled(config) -> None:
    if should_show_welcome_banner(config.banner) and \
            not config.get('quiet, False') and \
            not config.get('progress', False):
        show_welcome_banner()



from typing import Optional

from ...intel.signatures import SignatureSet
from ...logging import log
from ...util import vectorscan

from .matching import MatchEngineOptions, Matcher, BaseMatcherContext, \
        MatchWorkspace, MatchEngineCompilerOptions, Compiler


if not vectorscan.AVAILABLE:
    raise RuntimeError('Vectorscan is not available')


from ...util.vectorscan import VectorscanStreamScanner, VectorscanMatch, \
        VectorscanFlags, VectorscanDatabase, VectorscanScanTerminated, \
        VectorscanMode, vectorscan_compile, vectorscan_deserialize, \
        VectorscanPlatformInfo, VectorscanCpuFeatures, VectorscanTuneFamily, \
        VectorscanDatabaseIncompatible


class VectorscanMatcherContext(BaseMatcherContext):

    def __init__(self, matcher: Matcher):
        super().__init__(matcher)
        self.matched = None

    def _match_callback(self, match: VectorscanMatch) -> bool:
        if not self.matcher.signature_set.has_signature(match.identifier):
            return False
        self._record_match(
                identifier=match.identifier,
                matched=None
            )
        self.matched = True
        return False if self.matcher.match_all else True

    def process_chunk(
                self,
                chunk: bytes,
                start: bool = False,
                workspace: Optional[MatchWorkspace] = None
            ) -> bool:
        self.matched = False
        try:
            self.matcher.scanner.scan(chunk)
        except VectorscanScanTerminated:
            return True
        return self.matched

    def finalize_content(self) -> None:
        self.matcher.scanner.reset()

    def __enter__(self):
        self.matcher.scanner.set_callback(self._match_callback)
        return super().__enter__()

    def __exit__(self, exc_type, exc_value, traceback) -> None:
        pass


class VectorscanCompiler(Compiler):

    def __init__(self, generic: bool = False):
        self.generic = generic

    def compile(self, signature_set: SignatureSet) -> bytes:
        patterns = {
                signature.identifier: signature.rule
                for signature in signature_set.signatures.values()
            }
        pattern_count = len(patterns)
        log.info(
                f'Compiling {pattern_count} pattern(s) '
                'to vectorscan database (this may take a substantial '
                'amount of time depending on system performance)...'
            )
        flags = (
                VectorscanFlags.CASELESS |
                VectorscanFlags.SINGLEMATCH |
                VectorscanFlags.ALLOWEMPTY
            )
        platform_info = VectorscanPlatformInfo(
                VectorscanCpuFeatures.NONE,
                VectorscanTuneFamily.GENERIC
            ) if self.generic else None
        database = vectorscan_compile(
                patterns,
                mode=VectorscanMode.STREAM,
                flags=flags,
                platform_info=platform_info
            )
        log.info('Successfully compiled vectorscan database')
        return database

    def compile_serializable(self, signature_set: SignatureSet) -> bytes:
        database = self.compile(signature_set)
        return database.serialize()


class VectorscanMatcher(Matcher):

    def __init__(
                self,
                signature_set: SignatureSet,
                match_all: bool = False,
                database_source: Optional[bytes] = None,
                lazy: bool = False,
            ):
        self.signature_set = signature_set
        self.match_all = match_all
        self.database_source = database_source
        self.database = None
        self.scanner = None
        super().__init__(
                signature_set=signature_set,
                match_all=match_all,
                lazy=lazy
            )

    def _compile_database(self) -> VectorscanDatabase:
        compiler = VectorscanCompiler()
        return compiler.compile(self.signature_set)

    def _load_database(self, data) -> VectorscanDatabase:
        log.debug('Deserializing pre-compiled vectorscan database...')
        database = vectorscan_deserialize(data)
        log.debug('Successfully deserialized vectorscan database')
        return database

    def _initialize_database(self) -> VectorscanDatabase:
        if self.database_source:
            return self._load_database(self.database_source)
        else:
            return self._compile_database()

    def _prepare(self) -> None:
        log.debug('Preparing vectorscan matcher...')
        self.database = self._initialize_database()
        log.debug('Successfully prepared vectorscan matcher')

    def _prepare_thread(self) -> None:
        log.debug('Preparding thread-specific vectorscan scanner...')
        self.scanner = VectorscanStreamScanner(self.database)
        log.debug('Successfully prepared vectorscan scanner')

    def create_context(self) -> VectorscanMatcherContext:
        return VectorscanMatcherContext(
                self
            )


def create_compiler(options: MatchEngineCompilerOptions) -> VectorscanCompiler:
    return VectorscanCompiler(generic=options.generic)


def validate_database_source(source: bytes) -> bool:
    try:
        vectorscan_deserialize(source)
        return True
    except VectorscanDatabaseIncompatible as e:
        log.exception(e)
        return False


def create_matcher(options: MatchEngineOptions) -> VectorscanMatcher:
    return VectorscanMatcher(
            options.signature_set,
            match_all=options.match_all,
            database_source=options.database_source,
            lazy=options.lazy,
        )



import os

from typing import List, Optional, Dict
from email.message import EmailMessage
from email.headerregistry import Address

from ...wordpress.remediator import RemediationResult
from ...wordpress.identifier import FileType
from ...util.terminal import Color, escape, RESET
from ..email import Mailer
from ..reporting import ReportColumnEnum, Report, ReportManager, \
        ReportFormatEnum, ReportFormat, ReportColumn, ReportRecord, \
        ReportEmail, BaseHumanReadableWriter, \
        get_config_options, generate_report_email_html, generate_html_table, \
        REPORT_FORMAT_CSV, REPORT_FORMAT_TSV, REPORT_FORMAT_NULL_DELIMITED, \
        REPORT_FORMAT_LINE_DELIMITED
from ..context import CliContext
from ...util.encoding import bytes_to_str


class RemediationReportColumn(ReportColumnEnum):
    PATH = 'path', lambda record: os.fsdecode(record.result.path),
    STATUS = 'status', lambda record: record.get_status(),
    TYPE = 'type', lambda record: record.result.identity.type,
    SITE = 'site', \
        lambda record: os.fsdecode(
                record.result.identity.site.core_path
            ) \
        if record.result.identity.site is not None \
        else None
    TARGET_PATH = 'target_path', \
        lambda record: os.fsdecode(record.result.target_path),
    WORDPRESS_VERSION = 'wordpress_version', \
        lambda record: bytes_to_str(
                record.result.identity.site.get_version()
            ) \
        if record.result.identity.site is not None \
        else None
    EXTENSION_SLUG = 'extension_slug', \
        lambda record: record.result.identity.extension.slug \
        if record.result.identity.extension is not None else None
    EXTENSION_NAME = 'extension_name', \
        lambda record: record.result.identity.extension.get_name() \
        if record.result.identity.extension is not None else None
    EXTENSION_VERSION = 'extension_version', \
        lambda record: bytes_to_str(
                record.result.identity.extension.version
            ) \
        if record.result.identity.extension is not None else None


class HumanReadableWriter(BaseHumanReadableWriter):

    def format_record(self, record) -> str:
        result = record.result
        path_str = os.fsdecode(result.path)
        if result.remediated:
            green = escape(Color.GREEN)
            return f'{green}Successfully remediated {path_str}{RESET}'
        elif not result.known:
            yellow = escape(Color.YELLOW)
            return (
                    f'{yellow}Path at {path_str} is unknown and cannot'
                    f' be remediated{RESET}'
                )
        else:
            red = escape(Color.RED)
            return f'{red}Remediation for {path_str} failed{RESET}'


REPORT_FORMAT_HUMAN = ReportFormat(
        'human',
        lambda stream, columns: HumanReadableWriter(stream),
        allows_headers=False,
        allows_column_customization=False
    )


class RemediationReportFormat(ReportFormatEnum):
    CSV = REPORT_FORMAT_CSV
    TSV = REPORT_FORMAT_TSV
    NULL_DELIMITED = REPORT_FORMAT_NULL_DELIMITED
    LINE_DELIMITED = REPORT_FORMAT_LINE_DELIMITED
    HUMAN = REPORT_FORMAT_HUMAN


class RemediationReportRecord(ReportRecord):

    def __init__(self, result: RemediationResult):
        self.result = result

    def get_status(self) -> str:
        if self.result.remediated:
            return 'remediated'
        if self.result.identity.type == FileType.UNKNOWN:
            return 'unidentified'
        if not self.result.known:
            return 'unrecognized'
        return 'failed'


class RemediationCounts:

    def __init__(self):
        self.total = 0
        self.remediated = 0
        self.known = 0
        self.unknown = 0

    def add(self, result: RemediationResult):
        if result.remediated:
            self.remediated += 1
        if result.identity.type is FileType.UNKNOWN:
            self.unknown += 1
        else:
            self.known += 1
        self.total += 1

    @property
    def unsuccessful(self):
        return self.total - self.remediated


class RemediationReport(Report):

    def __init__(
                self,
                format: ReportFormat,
                columns: List[ReportColumn],
                email_addresses: List[str],
                mailer: Optional[Mailer],
                write_headers: bool = False,
                only_unremediated: bool = False
            ):
        super().__init__(
                format,
                columns,
                email_addresses,
                mailer,
                write_headers
            )
        self.only_unremediated = only_unremediated
        self.counts = RemediationCounts()

    def add_result(self, result: RemediationResult):
        self.counts.add(result)
        if self.only_unremediated and result.remediated:
            return
        self.write_record(
                RemediationReportRecord(
                    result
                )
            )

    def generate_email(
                self,
                recipient: Address,
                attachments: Dict[str, str],
                hostname: str
            ) -> EmailMessage:

        plain = 'Remediation Complete\n\n' \
                'Successfully Remediated Files: ' \
                f'{self.counts.remediated}\n\n' \
                f'Total Files Processed: {self.counts.total}\n\n' \

        results = {
                'Remediated': self.counts.remediated,
                'Failed': self.counts.unsuccessful,
                'Unknown Files': self.counts.unknown,
                'Total': self.counts.total
            }

        table = generate_html_table(results)

        document = generate_report_email_html(
                table,
                'Remediation Results',
                hostname
            )

        return ReportEmail(
                recipient=recipient,
                subject=f'Remediation Results for {hostname}',
                plain_content=plain,
                html_content=document.to_html()
            )


class RemediationReportManager(ReportManager):

    def __init__(self, context: CliContext):
        super().__init__(
                formats=RemediationReportFormat,
                columns=RemediationReportColumn,
                context=context,
                read_stdin=context.config.read_stdin,
                input_delimiter=context.config.path_separator,
                binary_input=True
            )
        self.only_unremediated = context.config.output_unremediated

    def _instantiate_report(
                self,
                format: ReportFormat,
                columns: List[ReportColumn],
                email_addresses: List[str],
                mailer: Optional[Mailer],
                write_headers: bool
            ) -> RemediationReport:
        return RemediationReport(
                format,
                columns,
                email_addresses,
                mailer,
                write_headers,
                self.only_unremediated
            )


REMEDIATION_REPORT_CONFIG_OPTIONS = get_config_options(
        RemediationReportFormat,
        RemediationReportColumn,
        default_format='human'
    )



from ctypes import Structure, POINTER, c_char_p, c_int, \
    c_void_p, c_uint, c_ulonglong, c_size_t, byref, string_at, CFUNCTYPE
from enum import IntFlag, IntEnum
from typing import Dict, Optional, Callable, Union, Any

from ..library import load_library, LibraryNotAvailableException
from ..encoding import bytes_to_str
from .. import signals

from .vectorscan import VectorscanException, \
    VectorscanLibraryNotAvailableException


try:
    hs = load_library('hs')
except LibraryNotAvailableException:
    raise VectorscanLibraryNotAvailableException('Failed to load libhs')


_hs_version = hs.hs_version
_hs_version.argtypes = []
_hs_version.restype = c_char_p
VERSION = bytes_to_str(_hs_version())
API_VERSION = ''.join(VERSION.split()[:1])


class _StructHsDatabase(Structure):
    pass


_hs_database_p = POINTER(_StructHsDatabase)


_hs_error = c_int


class _StructHsCompileError(Structure):
    _fields_ = [
            ('message', c_char_p),
            ('expression', c_int)
        ]


_hs_compile_error_p = POINTER(_StructHsCompileError)


class _StructHsPlatformInfo(Structure):
    _fields_ = [
            ('tune', c_uint),
            ('cpu_features', c_ulonglong),
            ('reserved1', c_ulonglong),
            ('reserved2', c_ulonglong)
        ]


_hs_platform_info_p = POINTER(_StructHsPlatformInfo)


_hs_compile_multi = hs.hs_compile_multi
_hs_compile_multi.argtypes = [
        POINTER(c_char_p),
        POINTER(c_uint),
        POINTER(c_uint),
        c_uint,
        c_uint,
        _hs_platform_info_p,
        POINTER(_hs_database_p),
        POINTER(_hs_compile_error_p)
    ]
_hs_compile_multi.restype = _hs_error


_hs_free_database = hs.hs_free_database
_hs_free_database.argtypes = [_hs_database_p]
_hs_free_database.restype = None


_hs_serialize_database = hs.hs_serialize_database
_hs_serialize_database.argtypes = [
        _hs_database_p,
        POINTER(c_char_p),
        POINTER(c_size_t)
    ]
_hs_serialize_database.restype = _hs_error


_hs_deserialize_database = hs.hs_deserialize_database
_hs_deserialize_database.argtypes = [
        c_char_p,
        c_size_t,
        POINTER(_hs_database_p)
    ]


class _StructHsScratch(Structure):
    pass


_hs_scratch_p = POINTER(_StructHsScratch)


_hs_alloc_scratch = hs.hs_alloc_scratch
_hs_alloc_scratch.argtypes = [_hs_database_p, POINTER(_hs_scratch_p)]
_hs_alloc_scratch.restype = _hs_error


_hs_free_scratch = hs.hs_free_scratch
_hs_free_scratch.argtypes = [_hs_scratch_p]
_hs_free_scratch.restype = None


_match_event_handler = CFUNCTYPE(
        c_int,
        c_uint,
        c_ulonglong,
        c_ulonglong,
        c_uint,
        c_void_p
    )


_hs_scan = hs.hs_scan
_hs_scan.argtypes = [
        _hs_database_p,
        c_char_p,
        c_uint,
        c_uint,
        _hs_scratch_p,
        _match_event_handler,
        c_void_p
    ]
_hs_scan.restype = _hs_error


class _StructHsStream(Structure):
    pass


_hs_stream_p = POINTER(_StructHsStream)


_hs_open_stream = hs.hs_open_stream
_hs_open_stream.argtypes = [
        _hs_database_p,
        c_uint,
        POINTER(_hs_stream_p)
    ]
_hs_open_stream.restype = _hs_error


_hs_scan_stream = hs.hs_scan_stream
_hs_scan_stream.argtypes = [
        _hs_stream_p,
        c_char_p,
        c_uint,
        c_uint,
        _hs_scratch_p,
        _match_event_handler,
        c_void_p
    ]
_hs_scan_stream.restype = _hs_error


_hs_reset_stream = hs.hs_reset_stream
_hs_reset_stream.argtypes = [
        _hs_stream_p,
        c_uint,
        _hs_scratch_p,
        _match_event_handler,
        c_void_p
    ]
_hs_reset_stream.restype = _hs_error


_hs_close_stream = hs.hs_close_stream
_hs_close_stream.argtypes = [
        _hs_stream_p,
        _hs_scratch_p,
        _match_event_handler,
        c_void_p
    ]
_hs_close_stream.restype = _hs_error


class VectorscanFlags(IntFlag):
    NONE = 0
    CASELESS = 1
    DOTALL = 2
    MULTILINE = 4
    SINGLEMATCH = 8
    ALLOWEMPTY = 16
    UTF8 = 32
    UCP = 64
    PREFILTER = 128
    LEFTMODE = 256
    COMBINATION = 512
    QUIET = 1024


class VectorscanMode(IntEnum):
    BLOCK = 1
    STREAM = 2
    VECTORED = 4


class VectorscanErrorType(IntEnum):
    SUCCESS = 0
    INVALID = -1
    NOMEM = -2
    SCAN_TERMINATED = -3
    COMPILER_ERROR = -4
    DB_VERSION_ERROR = -5
    DB_PLATFORM_ERROR = -6
    DB_MODE_ERROR = -7
    BAD_ALIGN = -8
    BAD_ALLOC = -9
    SCRATCH_IN_USE = -10
    ARCH_ERROR = -11
    INSUFFICIENT_SPACE = -12
    UNKNOWN_ERROR = -13


class VectorscanError(VectorscanException):

    def __init__(self, error: VectorscanErrorType):
        self.error = error


class VectorscanScanTerminated(VectorscanError):

    def __init__(self):
        super().__init__(VectorscanErrorType.SCAN_TERMINATED)


class VectorscanDatabaseIncompatible(VectorscanError):
    pass


DATABASE_COMPATIBILITY_ERRORS = [
        VectorscanErrorType.DB_VERSION_ERROR,
        VectorscanErrorType.DB_PLATFORM_ERROR,
        VectorscanErrorType.DB_MODE_ERROR
    ]


def _assert_success(error: Union[int, _hs_error]):
    try:
        if isinstance(error, _hs_error):
            error = _hs_error.value
        error = VectorscanErrorType(error)
    except ValueError:
        error = VectorscanErrorType.UNKNOWN_ERROR
    if error is not VectorscanErrorType.SUCCESS:
        if error is VectorscanErrorType.SCAN_TERMINATED:
            raise VectorscanScanTerminated()
        if error in DATABASE_COMPATIBILITY_ERRORS:
            raise VectorscanDatabaseIncompatible(error)
        raise VectorscanError(error)


class VectorscanCompilerError(VectorscanError):

    def __init__(self, message):
        super().__init__(VectorscanErrorType.COMPILER_ERROR)


def _assert_compilation_success(
            error: Union[int, _hs_error],
            compiler_error: _hs_compile_error_p
        ):
    try:
        _assert_success(error)
    except VectorscanError as e:
        if e.error is VectorscanErrorType.COMPILER_ERROR:
            raise VectorscanCompilerError(
                    compiler_error.contents.message.decode('utf-8')
                )
        else:
            raise


class VectorscanDatabase:

    def __init__(self, database: _hs_database_p):
        self._database = database

    def __del__(self) -> None:
        if self._database is not None:
            _hs_free_database(self._database)
            self._database = None

    def serialize(self) -> bytes:
        data = c_char_p()
        length = c_size_t()
        error = _hs_serialize_database(
                self._database,
                byref(data),
                byref(length)
            )
        _assert_success(error)
        return string_at(data, length.value)


class VectorscanScratch:

    def __init__(self, database: VectorscanDatabase):
        self._scratch = _hs_scratch_p()
        _hs_alloc_scratch(database._database, byref(self._scratch))

    def __del__(self) -> None:
        if self._scratch is not None:
            _hs_free_scratch(self._scratch)
            self._scratch = None


class VectorscanMatch:

    def __init__(
                self,
                identifier: int,
                start: int,
                end: int,
                context
            ):
        self.identifier = identifier
        self.start = start
        self.end = end
        self.context = context


VectorscanMatchCallback = Callable[[VectorscanMatch], bool]


def _default_match_callback(match: VectorscanMatch) -> bool:
    return False


def _wrap_match_callback(callback: VectorscanMatchCallback, context: Any):
    def wrapped_callback(
                identifier: int,
                start: int,
                end: int,
                _flags: int,
                c_context
            ) -> int:
        match = VectorscanMatch(
                identifier,
                start,
                end,
                context
            )
        should_terminate = callback(match)
        return 1 if should_terminate else 0
    return wrapped_callback


class VectorscanScanner:

    def __init__(
                self,
                database: VectorscanDatabase,
                scratch: Optional[VectorscanScratch] = None
            ):
        self.database = database
        self.scratch = scratch if scratch is not None \
            else VectorscanScratch(database)

    def _encode_data(self, data: Union[bytes, str]) -> bytes:
        if isinstance(data, str):
            data = data.encode('utf-8')
        return data


class VectorscanBlockScanner(VectorscanScanner):

    def __init__(
                self,
                database: VectorscanDatabase,
                scratch: Optional[VectorscanScratch] = None
            ):
        super().__init__(database, scratch)

    def scan(
                self,
                data: Union[bytes, str],
                callback: VectorscanMatchCallback,
                context: Optional[Any] = None
            ):
        data = self._encode_data(data)

        callback = _wrap_match_callback(callback, context)

        error = _hs_scan(
                self.database._database,
                c_char_p(data),
                c_uint(len(data)),
                c_uint(0),
                self.scratch._scratch,
                _match_event_handler(callback),
                c_void_p()
            )
        _assert_success(error)


class VectorscanStreamScanner(VectorscanScanner):

    def __init__(
                self,
                database: VectorscanDatabase,
                callback: VectorscanMatchCallback = _default_match_callback,
                context: Optional[Any] = None,
                scratch: Optional[VectorscanScratch] = None
            ):
        super().__init__(database, scratch)
        self._stream = None
        self.set_callback(callback, context)
        self._open_stream()

    def set_callback(
                self,
                callback: VectorscanMatchCallback,
                context: Optional[Any] = None
            ) -> None:
        self._callback = _match_event_handler(
                _wrap_match_callback(callback, context)
            )

    def _open_stream(self) -> None:
        stream = _hs_stream_p()
        error = _hs_open_stream(
                self.database._database,
                c_uint(0),
                byref(stream)
            )
        _assert_success(error)
        self._stream = stream

    def _close_stream(self) -> None:
        if self._stream is None:
            return
        error = _hs_close_stream(
                self._stream,
                self.scratch._scratch,
                self._callback,
                c_void_p()
            )
        _assert_success(error)
        self._stream = None

    def scan(
                self,
                data: Union[bytes, str]
            ) -> None:
        data = self._encode_data(data)
        error = _hs_scan_stream(
                self._stream,
                c_char_p(data),
                c_uint(len(data)),
                c_uint(0),
                self.scratch._scratch,
                self._callback,
                c_void_p()
            )
        _assert_success(error)

    def reset(self) -> None:
        if self._stream is None:
            self._open_stream()
        error = _hs_reset_stream(
                self._stream,
                c_uint(0),
                self.scratch._scratch,
                self._callback,
                c_void_p()
            )
        _assert_success(error)

    def __del__(self) -> None:
        self._close_stream()


class VectorscanCpuFeatures(IntFlag):
    NONE = 0
    AVX2 = 4
    AVX512 = 8
    AVX512VBMI = 16


class VectorscanTuneFamily(IntEnum):
    GENERIC = 0
    SNB = 1
    IVB = 2
    HSW = 3
    SLM = 4
    BDW = 5
    SKL = 6
    SKX = 7
    GLM = 8
    ICL = 9
    ICX = 10


class VectorscanPlatformInfo:

    def __init__(
                self,
                cpu_features: VectorscanCpuFeatures,
                tune_family: VectorscanTuneFamily
            ):
        self._platform_info = _StructHsPlatformInfo(
                c_uint(tune_family.value),
                c_ulonglong(cpu_features.value)
            )


def vectorscan_compile(
            patterns: Dict[int, str],
            mode: VectorscanMode = VectorscanMode.BLOCK,
            flags: VectorscanFlags = VectorscanFlags.NONE,
            platform_info: Optional[VectorscanPlatformInfo] = None
        ) -> VectorscanDatabase:
    database = _hs_database_p()
    compiler_error = _hs_compile_error_p()
    ids = [c_uint(id) for id in patterns.keys()]
    ids = (c_uint * len(ids))(*ids)
    expressions = [
            c_char_p(expression.encode('utf-8')) for expression
            in patterns.values()
        ]
    expressions = (c_char_p * len(expressions))(*expressions)
    c_flags = (c_uint * len(ids))()
    for i in range(0, len(ids)):
        c_flags[i] = c_uint(flags)
    signals.reset()
    platform_info_p = _hs_platform_info_p() if platform_info is None \
        else byref(platform_info._platform_info)
    error = _hs_compile_multi(
            expressions,
            c_flags,
            ids,
            c_uint(len(patterns)),
            c_uint(mode),
            platform_info_p,
            byref(database),
            byref(compiler_error)
        )
    signals.restore()
    _assert_compilation_success(error, compiler_error)
    return VectorscanDatabase(database)


def vectorscan_deserialize(data: bytes) -> VectorscanDatabase:
    _database = _hs_database_p()
    error = _hs_deserialize_database(
            c_char_p(data),
            c_size_t(len(data)),
            byref(_database)
        )
    _assert_success(error)
    return VectorscanDatabase(_database)



import argparse
import json
import os
from argparse import ArgumentParser, Namespace
from typing import Set, List, Dict, Any, Tuple

from wordfence.logging import log
from ..helper import Helper
from .config_items import ConfigItemDefinition, \
    CanonicalValueExtractorInterface, Context, ArgumentType, \
    not_set_token
from .base_config_definitions \
        import config_map as base_config_map
from ..subcommands import SubcommandDefinition

NAME = "Wordfence CLI"
DESCRIPTION = ("Multifunction commandline tool for Wordfence - "
               "use wordfence {subcommand} --help for additional"
               " information about subcommands"
               )
COMMAND = "wordfence"

valid_contexts: Set[Context] = {Context.ALL, Context.CLI}


class CliCanonicalValueExtractor(CanonicalValueExtractorInterface):
    def is_valid_source(self, source: Any) -> bool:
        return isinstance(source, Namespace)

    def get_canonical_value(self, definition: ConfigItemDefinition,
                            source: Namespace) -> Any:
        self.assert_is_valid_source(source)
        value = getattr(source, definition.property_name, not_set_token)

        # Unset repeatable options are returned as lists with a single
        # not_set_token entry. Other repeatable options with values include a
        # not_set_token entry in their list that should be discarded.
        if isinstance(value, List):
            while not_set_token in value:
                value.remove(not_set_token)
            # falsy if the list is empty, and only contained a not_set_token
            if not value:
                value = not_set_token
        return value

    def get_context(self) -> Context:
        return Context.CLI


def create_split_and_append_action(delimiter: str, value_type=None):

    if value_type is None:
        value_type = str

    class SplitAndAppend(argparse.Action):

        def __call__(
                    self,
                    parser: argparse.ArgumentParser,
                    namespace: argparse.Namespace,
                    values,
                    option_string=None
                ):
            items = getattr(namespace, self.dest, [])
            new_values = values.split(delimiter)
            items.extend(
                    [value_type(value) for value in new_values if value != '']
                )
            setattr(namespace, self.dest, items)

    return SplitAndAppend


def add_to_parser(target_parser,
                  config_definition: ConfigItemDefinition) -> None:
    if config_definition.context not in valid_contexts:
        log.warning(
            f"Config value {json.dumps(config_definition.name)} is not a valid"
            f" CLI argument. Should it be specified in the INI file instead?")
        return

    names: List[str] = [f"--{config_definition.name}"]
    if config_definition.short_name:
        names.append(f"-{config_definition.short_name}")

    # common arguments
    named_params: Dict[str, Any] = {
        'help': config_definition.description,
        'default': not_set_token,
        'action': 'store'
    }
    if config_definition.has_options_list():
        named_params['choices'] = config_definition.meta.valid_options

    # special handling
    if config_definition.is_flag():
        # store the opposite of the default boolean
        named_params['action'] = 'store_true'
        # adjust the provided help message
        defaults_to = f'true (--{config_definition.name})' if (
            config_definition.default) else \
            f'false (--no-{config_definition.name})'
        if config_definition.argument_type == ArgumentType.FLAG \
                and config_definition.default:
            named_params['help'] += (f' If not specified, defaults to '
                                     f'{defaults_to}.')
    elif config_definition.argument_type == ArgumentType.OPTION_REPEATABLE:
        named_params['action'] = 'append'
        named_params['default'] = [not_set_token]

    if config_definition.has_separator():
        named_params['default'] = [not_set_token]
        named_params['action'] = \
            create_split_and_append_action(
                    config_definition.meta.separator,
                    config_definition.get_value_type()
                )
    # store_true and store_false do not have the same options as other actions,
    # and will throw an error if type is specified
    elif not isinstance(named_params['action'], str) or \
            not named_params['action'].startswith('store_'):
        if config_definition.accepts_paths():
            named_params['type'] = os.fsencode
        else:
            named_params['type'] = config_definition.get_value_type()

    named_params['help'] = argparse.SUPPRESS

    target_parser.add_argument(*names, **named_params)

    # register the negation of a flag
    if config_definition.is_flag():
        named_params['action'] = 'store_false'
        names = [f"--no-{config_definition.name}"]
        named_params['help'] = argparse.SUPPRESS
        # set the value to override the un-prefixed command
        named_params['dest'] = config_definition.property_name
        target_parser.add_argument(*names, **named_params)


def add_definitions_to_parser(
            parser: ArgumentParser,
            definitions: Dict[str, ConfigItemDefinition]
        ) -> None:
    for definition in definitions.values():
        add_to_parser(parser, definition)


def get_cli_values(
            subcommand_definitions: Dict[str, SubcommandDefinition],
            helper: Helper
        ) -> Tuple[Namespace, List[str], ArgumentParser]:
    parser = ArgumentParser(
            prog=COMMAND,
            description=DESCRIPTION,
            add_help=False,
            usage=helper.generate_usage()
        )

    add_definitions_to_parser(parser, base_config_map)

    subparsers = parser.add_subparsers(title="Available Subcommands",
                                       dest="subcommand",
                                       metavar='')
    for subcommand_definition in subcommand_definitions.values():
        definitions = subcommand_definition.get_config_map()
        subparser = subparsers.add_parser(
                subcommand_definition.name,
                prog=subcommand_definition.name,
                add_help=False,
                usage=helper.generate_usage(),
            )
        add_definitions_to_parser(subparser, base_config_map)
        add_definitions_to_parser(subparser, definitions)

        for previous_name in subcommand_definition.previous_names:
            subparsers.add_parser(
                    previous_name,
                    prog=previous_name
                )

    cli_values, trailing_arguments = parser.parse_known_args()
    if '--' in trailing_arguments:
        if trailing_arguments[0] != '--':
            unknowns = trailing_arguments[0:trailing_arguments.index('--')]
            unknowns = ', '.join(map(lambda x: json.dumps(x), unknowns))
            raise ValueError(f"Encountered unknown command arguments: "
                             f"{unknowns}")
        trailing_arguments = trailing_arguments[1:]
    return cli_values, trailing_arguments, parser
